import pymadx as _pymadx
import pybdsim as _pybdsim
import matplotlib.pyplot as _plt
import numpy as _np
from os.path import isfile as _isfile
from matplotlib.backends.backend_pdf import PdfPages as _PdfPages
import datetime as _datetime

def MadxVsBDSIM(tfs, bdsim, survey=None, functions=None,
                postfunctions=None, figsize=(10, 5), saveAll=True, outputFileName=None):
    """
    Compares MadX and BDSIM optics variables.
    User must provide a tfsoptIn file or Tfsinstance and a BDSAscii file or instance.

    +-----------------+---------------------------------------------------------+
    | **Parameters**  | **Description**                                         |
    +-----------------+---------------------------------------------------------+
    | tfs             | Tfs file or pymadx.Data.Tfs instance.                   |
    +-----------------+---------------------------------------------------------+
    | bdsim           | Optics root file (from rebdsimOptics or rebdsim).       |
    +-----------------+---------------------------------------------------------+
    | survey          | BDSIM model survey.                                     |
    +-----------------+---------------------------------------------------------+
    | functions       | Hook for users to add their functions that are called   |
    |                 | immediately prior to the addition of the plot. Use a    |
    |                 | lambda function to add functions with arguments. Can    |
    |                 | be a function or a list of functions.                   |
    +-----------------+---------------------------------------------------------+
    | figsize         | Figure size for all figures - default is (12,5)         |
    +-----------------+---------------------------------------------------------+
    
    """

    _CheckFilesExist(tfs, bdsim, survey)

    fname = _pybdsim.Data.GetFileName(bdsim) # cache file name
    if fname == "":
        fname = "optics_report"

    tfsinst = _pymadx.Data.CheckItsTfs(tfs)
    bdsinst = _pybdsim.Data.CheckItsBDSAsciiData(bdsim, True)

    tfsheader = tfsinst.header
    tfsopt  = _GetTfsOptics(tfsinst)
    bdsopt  = _GetBDSIMOptics(bdsinst)

    if survey is None:
        survey = tfsinst

    figures = [PlotBetas(tfsopt, bdsopt, survey=survey,
                         functions=functions,
                         postfunctions=postfunctions,
                         figsize=figsize),
               PlotAlphas(tfsopt, bdsopt, survey=survey,
                          functions=functions,
                          postfunctions=postfunctions,
                          figsize=figsize),
               PlotDs(tfsopt, bdsopt, survey=survey,
                      functions=functions,
                      postfunctions=postfunctions,
                      figsize=figsize),
               PlotDps(tfsopt, bdsopt, survey=survey,
                       functions=functions,
                       postfunctions=postfunctions,
                       figsize=figsize),
               PlotSigmas(tfsopt, bdsopt, survey=survey,
                          functions=functions,
                          postfunctions=postfunctions,
                          figsize=figsize),
               PlotSigmasP(tfsopt, bdsopt, survey=survey,
                           functions=functions,
                           postfunctions=postfunctions,
                           figsize=figsize),
               PlotMeans(tfsopt, bdsopt, survey=survey,
                         functions=functions,
                         postfunctions=postfunctions,
                         figsize=figsize),
               PlotEmitt(tfsopt, bdsopt, tfsinst.header, survey=survey,
                         functions=functions,
                         postfunctions=postfunctions,
                         figsize=figsize),
               PlotNParticles(bdsopt, survey=survey,
                              functions=functions,
                              postfunctions=postfunctions,
                              figsize=figsize)]

    if saveAll:
        tfsname = repr(tfsinst)
        bdsname = repr(bdsinst)
        output_filename = "optics-report.pdf"
        if outputFileName is not None:
            output_filename = outputFileName
            if not output_filename.endswith('.pdf'):
                output_filename += ".pdf"
        else:
            output_filename = fname.replace('.root','')
            output_filename += ".pdf"
        # Should have a more descriptive name really.
        with _PdfPages(output_filename) as pdf:
            for figure in figures:
                pdf.savefig(figure)
            d = pdf.infodict()
            d['Title'] = "{} (TFS) VS {} (BDSIM) Optical Comparison".format(tfsname, bdsname)
            d['CreationDate'] = _datetime.datetime.today()

        print("Written ", output_filename)

def PrepareResiduals(tfs, bds, survey=None, verbose=False):
    """
    Filter the tfs and bds to provide data that will match the 
    BDSIM data in element name. 
    """
    _CheckFilesExist(tfs, bds, survey)
    tfsinst   = _pymadx.Data.CheckItsTfs(tfs)
    bdsinst   = _pybdsim.Data.CheckItsBDSAsciiData(bds) # works for root files too

    bdel    = bdsinst.orbit.elementName()
    bdslist = _np.array(bdel)
    tfsa  = tfsinst.GetColumn("NAME")

    keys = ['S', 'X', 'PX', 'Y', 'PY']

    tfsdata = {
        'S':tfsinst.GetColumn("S"),
        'X':tfsinst.GetColumn("X"),
        'PX':tfsinst.GetColumn("PX"),
        'Y':tfsinst.GetColumn("Y"),
        'PY':tfsinst.GetColumn("PY")
        }

    bdsdata = {
        'S':_np.array(bdsinst.orbit.s()),
        'X':_np.array(bdsinst.orbit.x()),
        'PX':_np.array(bdsinst.orbit.xp()),
        'Y':_np.array(bdsinst.orbit.y()),
        'PY':_np.array(bdsinst.orbit.yp())
        }

    reducedName = [_pybdsim._General.PrepareReducedName(name) for name in tfsa]
    rn = _np.array(reducedName)

    for n in keys:
        tfsdata[n] = tfsdata[n][_np.isin(rn,bdslist)]
        bdsdata[n] = bdsdata[n][_np.isin(bdslist,rn)]
        tfsdata[n] = _np.array(tfsdata[n])
        bdsdata[n] = _np.array(bdsdata[n])   

    return tfsdata,bdsdata 
 

def MadxVsBDSIMFromGMAD(tfs, gmad, outputfilename):
    """
    Runs the BDSIM model provided by the gmad file given, gets the
    optics, and then compares them with TFS.

    tfs - path to TFS file or a pymadx.Data.Tfs instance
    gmad - path to gmad file to be run.
    outputfilename - path of the output optics report file.

    """

    bdsimOptics = _pybdsim.Run.GetOpticsFromGMAD(gmad)
    # Use TFS for survey but perhaps one day can directly get the
    # model from the ROOT output (having modified the above
    # function..  Currently there's no interface for this.
    MadxVsBDSIM(tfs, bdsimOptics, survey=tfs, outputFileName=outputfilename)


def _GetBDSIMOptics(optics):
    '''
    Takes a BDSAscii instance.
    Return a dictionary of lists matching the variable with the list of values.
    '''
    
    optvars = {}
    for variable in optics.names:
        datum = getattr(optics, variable)()
        optvars[variable] = datum
    return optvars

def _GetTfsOptics(optics):
    '''
    Takes either Tfs instance.  Returns dictionary of lists.
    '''

    MADXOpticsVariables = frozenset(['NAME',
                                     'S',
                                     'BETX',
                                     'BETY',
                                     'ALFX',
                                     'ALFY',
                                     'DX',
                                     'DPX',
                                     'DY',
                                     'DPY',
                                     'DXBETA',
                                     'DPXBETA',
                                     'DYBETA',
                                     'DPYBETA',
                                     'SIGMAX',
                                     'SIGMAY',
                                     'SIGMAXP',
                                     'SIGMAYP',
                                     'X',
                                     'Y',
                                     'PX',
                                     'PY'])

    optvars = {}
    for variable in MADXOpticsVariables:
        optvars[variable] = optics.GetColumn(variable)
    return optvars

def _GetTfsOrbit(optics):
    '''
    Takes either Tfs instance.  Returns dictionary of lists.
    '''

    MADXOpticsVariables = frozenset(['S',
                                     'X',
                                     'Y',
                                     'PX',
                                     'PY'])

    optvars = {}
    for variable in MADXOpticsVariables:
        optvars[variable] = optics.GetColumn(variable)
    return optvars

def PlotBetas(tfsopt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    betaPlot = _plt.figure('Beta', figsize=figsize)
    _plt.plot(tfsopt['S'], tfsopt['BETX'], 'b', label=r'MADX $\beta_{x}$')
    _plt.plot(tfsopt['S'], tfsopt['BETY'], 'g', label=r'MADX $\beta_{y}$')

    #bds
    _plt.errorbar(bdsopt['S'], bdsopt['Beta_x'],
                  yerr=bdsopt['Sigma_Beta_x'],
                  label=r'BDSIM $\beta_{x}$' + ' ; N = ' + N,
                  marker='x',
                  ls = '',
                  color='b')

    _plt.errorbar(bdsopt['S'], bdsopt['Beta_y'],
                  yerr=bdsopt['Sigma_Beta_y'],
                  label=r'BDSIM $\beta_{y}$' + ' ; N = ' + N,
                  marker='x',
                  ls = '',
                  color='g')

    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\beta_{x,y}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    print("survey =", survey)

    _CallUserFigureFunctions(functions)
    _AddSurvey(betaPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    
    _plt.show(block=False)
    return betaPlot

def PlotAlphas(tfsopt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    alphaPlot = _plt.figure('Alpha', figsize=figsize)
    #tfs
    _plt.plot(tfsopt['S'], tfsopt['ALFX'], 'b', label=r'MADX $\alpha_{x}$')
    _plt.plot(tfsopt['S'], tfsopt['ALFY'], 'g', label=r'MADX $\alpha_{y}$')
    #bds
    _plt.errorbar(bdsopt['S'], bdsopt['Alpha_x'],
                  yerr=bdsopt['Sigma_Alpha_x'],
                  label=r'BDSIM $\alpha_{x}$' + ' ; N = ' + N,
                  fmt='b.', capsize=3)

    _plt.errorbar(bdsopt['S'], bdsopt['Alpha_y'],
                  yerr=bdsopt['Sigma_Alpha_y'],
                  label=r'BDSIM $\alpha_{y}$' + ' ; N = ' + N,
                  fmt='g.', capsize=3)

    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\alpha_{x,y}$')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(alphaPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    
    _plt.show(block=False)
    return alphaPlot

def PlotDs(tfsopt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    dispPlot = _plt.figure('Dispersion', figsize=figsize)
    #tfs
    _plt.plot(tfsopt['S'], tfsopt['DXBETA'], 'b', label=r'MADX $D_{x}$')
    _plt.plot(tfsopt['S'], tfsopt['DYBETA'], 'g', label=r'MADX $D_{y}$')
    #bds
    _plt.errorbar(bdsopt['S'], bdsopt['Disp_x'],
                  yerr=bdsopt['Sigma_Disp_x'],
                  label=r'BDSIM $D_{x}$' + ' ; N = ' + N,
                  fmt='b.', capsize=3)

    _plt.errorbar(bdsopt['S'], bdsopt['Disp_y'],
                  yerr=bdsopt['Sigma_Disp_y'],
                  label=r'BDSIM $D_{y}$' + ' ; N = ' + N,
                  fmt='g.', capsize=3)

    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$D_{x,y} / m$')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(dispPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    
    _plt.show(block=False)
    return dispPlot

def PlotDps(tfsopt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    dispPPlot = _plt.figure('Momentum_Dispersion', figsize=figsize)
    #tfs
    _plt.plot(tfsopt['S'], tfsopt['DPXBETA'], 'b', label=r'MADX $D_{p_{x}}$')
    _plt.plot(tfsopt['S'], tfsopt['DPYBETA'], 'g', label=r'MADX $D_{p_{y}}$')
    #bds
    _plt.errorbar(bdsopt['S'], bdsopt['Disp_xp'],
                  yerr=bdsopt['Sigma_Disp_xp'],
                  label=r'BDSIM $D_{p_{x}}$' + ' ; N = ' + N,
                  fmt='b.', capsize=3)

    _plt.errorbar(bdsopt['S'], bdsopt['Disp_yp'],
                  yerr=bdsopt['Sigma_Disp_yp'],
                  label=r'BDSIM $D_{p_{y}}$' + ' ; N = ' + N,
                  fmt='g.', capsize=3)

    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$D_{p_{x},p_{y}}$ / rad')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(dispPPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    
    _plt.show(block=False)
    return dispPPlot

def PlotEmitt(tfsopt, bdsopt, header, survey=None, functions=None, postfunctions=None, figsize=(12, 5)):
    N = str(int(bdsopt['Npart'][0]))  # number of primaries.
    emittPlot = _plt.figure('Emittance', figsize=figsize)
    ex = header['EX'] * _np.ones(len(tfsopt['S']))
    ey = header['EY'] * _np.ones(len(tfsopt['S']))

    # tfs
    _plt.plot(tfsopt['S'], ex, 'b', label=r'MADX $E_{x}$')
    _plt.plot(tfsopt['S'], ey, 'g', label=r'MADX $E_{x}$')
    # bds
    _plt.errorbar(bdsopt['S'], bdsopt['Emitt_x'],
                  yerr=bdsopt['Sigma_Emitt_x'],
                  label=r'BDSIM $E_{x}$' + ' ; N = ' + N,
                  fmt='b.', capsize=3)

    _plt.errorbar(bdsopt['S'], bdsopt['Emitt_y'],
                  yerr=bdsopt['Sigma_Emitt_y'],
                  label=r'BDSIM $E_{y}$' + ' ; N = ' + N,
                  fmt='g.', capsize=3)

    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$E_{x,y} / m$')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(emittPlot, survey)
    _CallUserFigureFunctions(postfunctions)

    _plt.show(block=False)
    return emittPlot

def PlotSigmas(tfsopt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    sigmaPlot = _plt.figure('Sigma', figsize=figsize)
    #tfs
    _plt.plot(tfsopt['S'], tfsopt['SIGMAX'], 'b', label=r'MADX $\sigma_{x}$')
    _plt.plot(tfsopt['S'], tfsopt['SIGMAY'], 'g', label=r'MADX $\sigma_{y}$')
    #bds
    _plt.errorbar(bdsopt['S'],
                  bdsopt['Sigma_x'],
                  yerr=bdsopt['Sigma_Sigma_x'],
                  label=r'BDSIM $\sigma_{x}$' + ' ; N = ' + N,
                  fmt='b.', capsize=3)

    _plt.errorbar(bdsopt['S'],
                  bdsopt['Sigma_y'],
                  yerr=bdsopt['Sigma_Sigma_y'],
                  label=r'BDSIM $\sigma_{y}$' + ' ; N = ' + N,
                  fmt='g.', capsize=3)
    
    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\sigma_{x,y}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(sigmaPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    
    _plt.show(block=False)
    return sigmaPlot

def PlotSigmasP(tfsopt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    sigmaPPlot = _plt.figure('SigmaP', figsize=figsize)
    #tfs
    _plt.plot(tfsopt['S'], tfsopt['SIGMAXP'], 'b', label=r'MADX $\sigma_{xp}$')
    _plt.plot(tfsopt['S'], tfsopt['SIGMAYP'], 'g', label=r'MADX $\sigma_{yp}$')
    #bds
    _plt.errorbar(bdsopt['S'],
                  bdsopt['Sigma_xp'],
                  yerr=bdsopt['Sigma_Sigma_xp'],
                  label=r'BDSIM $\sigma_{xp}$' + ' ; N = ' + N,
                  fmt='b.', capsize=3)

    _plt.errorbar(bdsopt['S'],
                  bdsopt['Sigma_yp'],
                  yerr=bdsopt['Sigma_Sigma_yp'],
                  label=r'BDSIM $\sigma_{yp}$' + ' ; N = ' + N,
                  fmt='g.', capsize=3)
    
    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\sigma_{xp,yp}$ / rad')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(sigmaPPlot, survey)
    _CallUserFigureFunctions(postfunctions)

    _plt.show(block=False)
    return sigmaPPlot

def PlotMeans(tfsopt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    meanPlot = _plt.figure('Mean', figsize=figsize)
    #tfs
    _plt.plot(tfsopt['S'], tfsopt['X'], 'b', label=r'MADX $\bar{x}$')
    _plt.plot(tfsopt['S'], tfsopt['Y'], 'g', label=r'MADX $\bar{y}$')

    #bdsim
    _plt.errorbar(bdsopt['S'], bdsopt['Mean_x'],
                  yerr=bdsopt['Sigma_Mean_x'],
                  label=r'BDSIM $\bar{x}$' + ' ; N = ' + N,
                  fmt='b.', capsize=3)

    _plt.errorbar(bdsopt['S'], bdsopt['Mean_y'],
                  yerr=bdsopt['Sigma_Mean_y'],
                  label=r'BDSIM $\bar{y}$' + ' ; N = ' + N,
                  fmt='g.', capsize=3)

    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\bar{x}, \bar{y}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(meanPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    
    _plt.show(block=False)
    return meanPlot

def PlotNParticles(bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12, 5)):
    npartPlot = _plt.figure('NParticles', figsize)

    _plt.plot(bdsopt['S'],bdsopt['Npart'], 'k-', label='BDSIM N Particles')
    _plt.plot(bdsopt['S'],bdsopt['Npart'], 'k.')
    axes = _plt.gcf().gca()
    axes.set_ylabel(r'N Particles')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(npartPlot, survey)
    _CallUserFigureFunctions(postfunctions)

    _plt.show(block=False)
    return npartPlot

def MadxVsBDSIMOrbit(tfs, bds, survey=None, functions=None, postfunctions=None, figsize=(12,5)):
    """
    Plot both the BDSIM orbit and MADX orbit (mean x,y).

    tfs - either file name or pymadx.Data.Tfs instance
    bds - filename or BDSAsciiData instance - rebdsimOrbit, rebdsimOptics output files
    """
    _CheckFilesExist(tfs, bds, survey)
    tfsopt = _pymadx.Data.CheckItsTfs(tfs)
    bdsopt = _pybdsim.Data.CheckItsBDSAsciiData(bds)

    if survey is None:
        survey = tfsopt

    orbitPlot = _plt.figure('Orbit', figsize=figsize)
    
    #tfs
    tfs_s = tfsopt.GetColumn('S')
    tfs_x = tfsopt.GetColumn('X')
    tfs_y = tfsopt.GetColumn('Y')
    _plt.plot(tfs_s, tfs_x, 'b', label=r'MADX $\bar{x}$')
    _plt.plot(tfs_s, tfs_y, 'g', label=r'MADX $\bar{y}$')

    #bdsim
    if type(bdsopt) == _pybdsim.Data.RebdsimFile:
        if (hasattr(bdsopt, "orbit")):
            bds_s = bdsopt.orbit.s()
            bds_x = bdsopt.orbit.x()
            bds_y = bdsopt.orbit.y()
        else:
            raise ValueError("No orbit in file")
    elif type(bdsopt) == _pybdsim.Data.BDSAsciiData:
        # ironically this will come if it's a rebdsim optics output
        bds_s = bdsopt.GetColumn('S')
        bds_x = bdsopt.GetColumn('Mean_x')
        bds_y = bdsopt.GetColumn('Mean_y')
    _plt.plot(bds_s, bds_x, 'b.', label='BDSIM x')
    #_plt.plot(bds_s, bds_x, 'b-', alpha=0.4)
    _plt.plot(bds_s, bds_y, 'g.', label='BDSIM y')
    #_plt.plot(bds_s, bds_y, 'g-', alpha=0.4)
    
    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\bar{x}, \bar{y}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    _CallUserFigureFunctions(functions)
    _AddSurvey(orbitPlot, survey)
    _CallUserFigureFunctions(postfunctions)

    _plt.show(block=False)
    return orbitPlot

def MadxVsBDSIMOrbitResiduals(tfs, bds, survey=None, functions=None, postfunctions=None, verbose=False, figsize=(12,5)):
    _CheckFilesExist(tfs, bds, survey)
    tfsinst   = _pymadx.Data.CheckItsTfs(tfs)
    bdsinst   = _pybdsim.Data.CheckItsBDSAsciiData(bds)
    tfsd = PrepareResiduals(tfs, bds)
    tdata = tfsd[0]
    bdata = tfsd[1]

    if survey is None:
        survey = tfs

    dx  = tdata['X']  - bdata['X']
    dxp = tdata['PX'] - bdata['PX']
    dy  = tdata['Y']  - bdata['Y']
    dyp = tdata['PY'] - bdata['PY']

    s   = tdata['S']

    orbRes = _plt.figure('OrbitResiduals', figsize=figsize)
    _plt.plot(s, abs(dx),  '.', label='x')
    _plt.plot(s, abs(dxp), '.', label='xp')
    _plt.plot(s, abs(dy),  '.', label='y')
    _plt.plot(s, abs(dyp), '.', label='yp')
    _plt.legend()
    _plt.xlabel('S (m)')
    _plt.ylabel('|residual| (m, rad)')
    _plt.yscale('log')

    _CallUserFigureFunctions(functions)
    _AddSurvey(orbRes, survey)

    _plt.show(block=False)
    return orbRes

def _AddSurvey(figure, survey):
    if survey is None:
        return
    if isinstance(survey, str): # If BDSIM ASCII survey file
        if survey.split(".")[-1] == 'dat':
            _pybdsim.Plot.AddMachineLatticeFromSurveyToFigure(figure,survey)
    # If BDSIM ASCII survey instance
    elif isinstance(survey, _pybdsim.Data.BDSAsciiData):
        _pymadx.Plot.AddMachineLatticeToFigure(figure,survey)
    elif isinstance(survey, _pymadx.Data.Tfs): # If TFS
        _pymadx.Plot.AddMachineLatticeToFigure(figure,survey)
    # if a (BDSIM) ROOT file
    elif _pybdsim._General.IsROOTFile(survey):
        pass

def _ProcessInput(tfsOptics, bdsimOptics):

    if not isinstance(tfsOptics, (_pymadx.Data.Tfs, str)):
        raise TypeError("tfsOptics should be either a path to a tfs file or "
                        "a pymadx.Data.Tfs instance!")
    if not isinstance(bdsimOptics, _pybdsim.Data.BDSAsciiData):
        raise TypeError("bdsimOptics should be either be a path to a "
                        "BDSAsciiData file or a pybdsim.Data.BDSAsciiData "
                        "instance")

    if isinstance(tfsOptics, str):
        tfsOptics = _pymadx.Data.Tfs(tfsOptics)
    if isinstance(tfsOptics, str):
        bdsimOptics = _pybdsim.Data.Load(bdsimOptics)

    return tfsOptics, bdsimOptics

def _CheckFilesExist(tfs, bdsim, survey):
    '''
    Otherwise such errors are too cryptic.
    '''
    if isinstance(tfs, str):
        if not _isfile(tfs):
            raise IOError("File not found: ", tfs)
    if isinstance(bdsim, str) and not _isfile(bdsim):
        raise IOError("File not found: ", bdsim)
    if isinstance(survey, str) and not _isfile(survey):
        raise IOError("File not found: ", survey)


def _CallUserFigureFunctions(functions):
    if isinstance(functions, list):
        for function in functions:
            if callable(function):
                function()
    elif callable(functions):
        functions()



from typing import Optional as _Optional
from typing import Dict as _Dict
from typing import Tuple as _Tuple
from typing import MutableMapping as _MutableMapping
from typing import Callable as _Callable
from typing import TYPE_CHECKING as _TYPE_CHECKING
import os as _os
import re as _re
import logging as _logging
import cpymad.madx as _madx
import yaml as _yaml
import pandas as _pd
import numpy as _np
from mergedeep import merge as _merge

import pybdsim.Builder as _Builder
import pybdsim.Beam as _Beam
import pybdsim.Options as _Options

# commented out as seems to be another dependency... don't know about it... only for type checking
#if _TYPE_CHECKING:
#    from pymask.madxp import Madxp

BDSIM_MAD_CONVENTION = {
    'apertype': 'apertureType',
    'circle': 'circular',
    'octagon': 'octagonal',
    'rectangle': 'rectangular',
    'collimator': 'rcol',
    'rcollimator': 'rcol',
    'electron': 'e-',
}

BDSIM_RESERVED_NAME = {
    'ms': 'ms.'
}

BDSIM_MADX_ELEMENTS_ARGUMENTS = ['l',
                                 'k1',
                                 'k1s',
                                 'k2',
                                 'k2s',
                                 'k3',
                                 'ks',
                                 'angle',
                                 'tilt',
                                 'kick',
                                 'hkick',
                                 'vkick',
                                 'knl',
                                 'ksl',
                                 'apertype',
                                 'aperture',
                                 'aper_offset']

BDSIM_ELEMENTS_TO_GENERATE = ['drift', 'sbend', 'rbend', 'quadrupole', 'sextupole', 'octupole',
                              'decapole', 'hkicker', 'vkicker', 'tkicker', 'kicker',
                              'solenoid', 'rcol', 'ecol', 'jcol', 'multipole', 'thinmultipole', 'rfcavity']

MAD_ELEMENTS_TO_DROP_IF_ZERO_LENGTH = ['marker', 'monitor', 'instrument', 'placeholder', 'rcollimator', 'cavity']


def _log_component_info(name: str, parent_name: str, base_name: str, max_length: int = 25) -> str:
    return ' ' * 5 + name.ljust(max_length, ' ') + parent_name.ljust(max_length, ' ') + base_name.ljust(max_length, ' ')


class CPyMad2Gmad:
    def __init__(self,
                 madx_instance,
                 model: _Optional[_Dict] = None,
                 aperture_model: _Callable[[str], str] = None,
                 model_path: str = '.',
                 model_file: str = 'model.yml',
                 log_file: str = 'model.log',
                 ):
        # Read main configuration file
        self.logging = _logging.getLogger()
        handler = _logging.FileHandler(log_file, mode='w')
        handler.setLevel(_logging.INFO)
        self.logging.addHandler(handler)
        self.logging.info("Reading the model configuration file.")
        with open(_os.path.join(model_path, model_file)) as file:
            _ = _yaml.full_load(file)

        # Build the complete model dictionary
        self.model = {}

        # need to check if there are some modules in the yml file
        if _['builder']['modules']:
            for module in _['builder']['modules']:
                with open(_os.path.join(
                        model_path,
                        _['builder']['config']['geometries_path'],
                        'modules',
                        module,
                        'data.yml')
                ) as file:
                    self.model = _merge(self.model, _yaml.full_load(file))
                    self.logging.info(f"Merged data from module '{module}'.")

        self.model = _merge(self.model, _)
        self.model = _merge(self.model, model or {})

        # Add quotes for BDSim options
        for k, v in self.model['options'].items():
            if isinstance(v, str) and '*' not in v:  # Physical quantity with units should not be quoted
                self.model['options'][k] = '"' + v + '"'

        if 'sequence' in self.model.keys():
            # Compile the regular expressions in the model
            self.logging.info("Compiling regular expressions...")
            for r in list(self.model['sequence'].keys()):
                self.model['sequence'][_re.compile(r)] = self.model['sequence'].pop(r)
            self.logging.info("... done.")

        # Aperture model
        self.aperture = aperture_model

        # MAD-X
        self.madx = madx_instance
        self.madx_beam = self.madx.sequence[self.model['builder']['sequence']].beam
        self.madx_sequence = self.madx.sequence[self.model['builder']['sequence']]
        self.madx_expanded_element_positions = self.madx_sequence.expanded_element_positions()
        self.madx_expanded_element_names = self.madx_sequence.expanded_element_names()

        def build_component(element, i, position, level):
            return {
                'NAME': BDSIM_RESERVED_NAME.get(element.name, element.name),
                'PARENT': BDSIM_MAD_CONVENTION.get(element.parent.name, element.parent.name),
                'BASE_PARENT': BDSIM_MAD_CONVENTION.get(element.base_type.name, element.base_type.name),
                'MAD_ELEMENT': element,
                'IS_ELEMENT': element is e,
                'IS_IMPLICIT_DRIFT': level - bottom == 0 and element.parent.name == 'drift',
                'LEVEL': level - bottom,
                'L': element.l,
                'AT': position,
                'ID': i,
                'USED': 1,
            }

        def build_component_recursive(i, element, position, level=0):
            nonlocal bottom
            if element.parent.name != element.name:
                if element.name not in components:
                    components[element.name] = build_component(element, i, position, level)
                    build_component_recursive(i, element.parent, position, level - 1)
                else:
                    components[element.name]['USED'] += 1
            else:
                bottom = level

        components = {}
        self.logging.info("Extracting all elements of the sequence to the dataframe...")
        for i, e, p in self._madx_subsequence():
            bottom = 0
            build_component_recursive(i, e, p)
        self.components = _pd.DataFrame(components.values())
        self.components.sort_values(by=['IS_IMPLICIT_DRIFT', 'IS_ELEMENT', 'BASE_PARENT', 'LEVEL'], inplace=True)
        self.logging.info("... done.")

        self.logging.info("Starting the conversion to BDSIM elements...")
        self.components['BDSIM'] = self.components.apply(self._build_bdsim_component, axis=1)
        self.components['BDSIM_TYPE'] = self.components.apply(lambda _: getattr(_['BDSIM'], 'category', None), axis=1)
        self.components.set_index("NAME", inplace=True)

        used = self.components['USED'].copy()
        for i, _ in self.components.iterrows():
            if _['BDSIM'] is None and _['PARENT'] in used:
                used.loc[_['PARENT']] -= 1
        self.components['USED'] = used
        keep = self.components['USED'] > 0

        self.logging.info("Dropping the following unused components:")
        for _ in keep[~keep].index:
            self.logging.info(' ' * 5 + _)
        self.components = self.components[keep]
        self.logging.info("... done.")

    @property
    def _madx_subsequence(self):
        def _iterator():
            mad_sequence = self.madx.sequence[self.model['builder']['sequence']]
            expanded_elements = mad_sequence.expanded_elements
            expanded_element_positions = mad_sequence.expanded_element_positions()
            idx1 = expanded_elements.index(self.model['builder']['from'] or expanded_elements[0].name)
            idx2 = expanded_elements.index(self.model['builder']['to'] or expanded_elements[-1].name)
            for i in range(idx1, idx2 + 1):
                element = expanded_elements[i]
                if element.base_type.name not in MAD_ELEMENTS_TO_DROP_IF_ZERO_LENGTH or element.get('l', 0.0) > 0.0:
                    yield i, element, expanded_element_positions[i]
                else:
                    self.logging.info(_log_component_info(element.name, element.parent.name, element.base_type.name)
                                      + f" - Skipping element because it should be dropped and its length is zero.")

        return _iterator

    def _get_model_properties_for_element(self, element_name: str) -> _Tuple[_Optional[str], dict]:
        properties = {}
        element_type = None
        for regex, data in self.model['sequence'].items():
            if regex.match(element_name):
                properties = _merge(properties, data['properties'])
                element_type: _Optional[str] = data.get('type', None)
        return element_type, properties

    def _build_bdsim_component_properties(self, element: _pd.Series) -> dict:
        mad_element = element['MAD_ELEMENT']
        mad_parent = mad_element.parent
        bdsim_properties = {}
        for k, v in mad_element.items():
            if k in BDSIM_MADX_ELEMENTS_ARGUMENTS:
                if isinstance(v, _madx.madx.ArrayAttribute):
                    if list(mad_parent.get(k)) != list(v):
                        bdsim_properties[BDSIM_MAD_CONVENTION.get(k, k)] = v
                else:
                    if mad_parent.get(k) != v:
                        bdsim_properties[BDSIM_MAD_CONVENTION.get(k, k)] = v
        if not bool(bdsim_properties):
            bdsim_properties = {'l': mad_element.l}
            self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                              + f" - Adding the length in the definition because bare aliases are not allowed.")

        return self._generate_apertures(element, self._adjust_bdsim_component_properties(element, bdsim_properties))

    def _generate_apertures(self, element: _pd.Series, bdsim_properties: dict) -> dict:
        """Generate missing apertures"""
        if element['IS_ELEMENT'] \
                and 'apertureType' not in bdsim_properties \
                and 'xsize' not in bdsim_properties \
                and 'ysize' not in bdsim_properties:
            self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                              + f" - Setting the aperture from the aperture model for this element.")
            if self.aperture:
                if element['NAME'] in self.aperture.keys():
                    apertype, aper1, aper2, aper3, aper4 = self.aperture[element['NAME']]
                else:
                    print(f"Attention, element {element['NAME']} not in aperture model")
                    apertype = "elliptical"
                    aper1 = 0.073
                    aper2 = 0.035
                    aper3 = 0.1
                    aper4 = 0.1
            else:
                apertype = "elliptical"
                aper1 = 0.073
                aper2 = 0.035
                aper3 = 0.1
                aper4 = 0.1
            if element['BASE_PARENT'] == 'rcol':
                bdsim_properties['xsize'] = aper1
                bdsim_properties['ysize'] = aper2
            else:
                bdsim_properties['apertureType'] = apertype
                bdsim_properties['aper1'] = aper1
                bdsim_properties['aper2'] = aper2
                bdsim_properties['aper3'] = aper3
                bdsim_properties['aper4'] = aper4

        return bdsim_properties

    def _adjust_bdsim_component_properties(self, element: _pd.Series, bdsim_properties: dict) -> dict:
        """For this method, explicit is better than implicit!"""

        def convert_aperture_value_definitions():
            k = 'aperture'
            if k in bdsim_properties:
                if len(bdsim_properties[k]) >= 1:
                    bdsim_properties['aper1'] = bdsim_properties[k][0]
                if len(bdsim_properties[k]) >= 2:
                    bdsim_properties['aper2'] = bdsim_properties[k][1]
                if len(bdsim_properties[k]) >= 3:
                    bdsim_properties['aper3'] = bdsim_properties[k][2]
                if len(bdsim_properties[k]) >= 4:
                    bdsim_properties['aper4'] = bdsim_properties[k][3]

        def set_collimator_gap_definitions():
            if element['BASE_PARENT'] == 'rcol':
                if 'aperture' in bdsim_properties:
                    bdsim_properties['xsize'] = bdsim_properties['aper1']
                    bdsim_properties['ysize'] = bdsim_properties['aper2']
                    if 'aper_offset' in bdsim_properties:
                        bdsim_properties['offsetX'] = bdsim_properties['aper_offset'][0]
                        bdsim_properties['offsetY'] = bdsim_properties['aper_offset'][1]
                        del bdsim_properties['aper_offset']
                    del bdsim_properties['aper1']
                    del bdsim_properties['aper2']
                    del bdsim_properties['aper3']
                    del bdsim_properties['aper4']
                    del bdsim_properties['apertureType']
                    self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                                      + f" - Setting the collimator opening: "
                                        f"xsize={bdsim_properties['xsize']}, ysize={bdsim_properties['ysize']}, "
                                        f"offsetX={bdsim_properties.get('offsetX', 0.0)}, "
                                        f"offsetY={bdsim_properties.get('offsetY', 0.0)}.")
                elif element['IS_ELEMENT']:
                    bdsim_properties['xsize'] = 0.05
                    bdsim_properties['ysize'] = 0.05
                    self.logging.warning(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                                         + f" - Setting the collimator opening with default values: "
                                           f"xsize={bdsim_properties['xsize']}, ysize={bdsim_properties['ysize']}.")

        def convert_aperture_types():
            k = 'apertureType'
            if k in bdsim_properties:
                bdsim_properties[k] = BDSIM_MAD_CONVENTION.get(bdsim_properties[k], bdsim_properties[k])

            k = 'apertureType'
            v = 'octagonal'
            if bdsim_properties.get(k) == v:
                # https://indico.cern.ch/event/379692/contributions/1804923/subcontributions/156446/attachments/757501/1039118/2105-03-18_HSS_meeting_rev.pdf
                aper3_mad = bdsim_properties['aper3']
                aper4_mad = bdsim_properties['aper4']
                bdsim_properties['aper3'] = bdsim_properties['aper2'] / _np.tan(aper4_mad)
                bdsim_properties['aper4'] = bdsim_properties['aper1'] * _np.tan(aper3_mad)

        def clean_apertures():
            k = 'aperture'
            if k in bdsim_properties:
                del bdsim_properties[k]

            k = ('aper1', 'aper2', 'aper3', 'aper4')
            for _ in k:
                threshold = self.model['builder']['config']['apertures']['remove_above_value']
                if bdsim_properties.get(_, 0.0) > threshold:
                    self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                                      + f" - Dropping the key {_} because it exceeds the treshold value {threshold}.")
                    del bdsim_properties[_]

            if 'aper1' not in bdsim_properties and 'aper2' not in bdsim_properties and 'aper3' not in bdsim_properties \
                    and 'aper4' not in bdsim_properties and 'apertureType' in bdsim_properties:
                del bdsim_properties['apertureType']
                self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                                  + f" - Dropping `apertureType` because it has no definition of aperture values.")

        def process_kicker_strengths():
            if element['BASE_PARENT'] == 'hkicker':
                if 'kick' in bdsim_properties:
                    if 'hkick' in bdsim_properties:
                        bdsim_properties['hkick'] += bdsim_properties['kick']
                    else:
                        bdsim_properties['hkick'] = bdsim_properties['kick']
                    del bdsim_properties['kick']
            if element['BASE_PARENT'] == 'vkicker':
                if 'kick' in bdsim_properties:
                    if 'vkick' in bdsim_properties:
                        bdsim_properties['vkick'] += bdsim_properties['kick']
                    else:
                        bdsim_properties['vkick'] = bdsim_properties['kick']
                    del bdsim_properties['kick']

        def process_multipolar_components():
            for _ in ('knl', 'ksl'):
                if _ in bdsim_properties:
                    bdsim_properties[_] = tuple(bdsim_properties[_])

        convert_aperture_value_definitions()
        set_collimator_gap_definitions()
        convert_aperture_types()
        clean_apertures()
        process_kicker_strengths()
        process_multipolar_components()

        return bdsim_properties

    def _generate_bdsim_component(self, element: _pd.Series, bdsim_properties: _MutableMapping):
        """
        Generate the Pybdsim component from its adjusted properties.

        Args:
            element:
            bdsim_properties:

        Returns:

        """
        parent_name = element['PARENT']
        if element['L'] == 0:
            if element['BASE_PARENT'] == 'multipole':
                element['LEVEL'] += 1

            if element['BASE_PARENT'] == 'multipole' \
                    and all(m == 0 for m in bdsim_properties.get('knl', ()) + bdsim_properties.get('ksl', ())) \
                    and element['LEVEL'] == 0:
                self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                                  + f" - Dropping multipole because its length is zero and it has no field.")
                return None
            if element['BASE_PARENT'] in ('kicker', 'hkicker', 'vkicker', 'tkicker') \
                    and bdsim_properties.get('hkick', 0.0) == 0 \
                    and bdsim_properties.get('vkick', 0.0) == 0 \
                    and element['LEVEL'] == 0:
                self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                                  + f" - Dropping kicker because its length is zero and it has no field.")
                return None
            # if :  # Means it is a base type
            #     parent_name = 'thinmultipole'
            #     logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
            #                  + f" - Converting to thinmultipole because it is a zero-length multipole.")

        if element['PARENT'] not in self.components['NAME'].values:
            if element['PARENT'] not in BDSIM_ELEMENTS_TO_GENERATE:
                parent_name = 'drift'
                self.logging.info(_log_component_info(element['NAME'], element['PARENT'], element['BASE_PARENT'])
                                  + f" - Converting to a drift because {element['PARENT']} is not supported.")
            bdsim_component = _Builder.Element(
                name=element['NAME'],
                category=parent_name,
                isMultipole='knl' in bdsim_properties or 'ksl' in bdsim_properties,
                **bdsim_properties)
        else:
            bdsim_component = _Builder.Element.from_element(
                name=element['NAME'],
                parent_element_name=parent_name,
                isMultipole='knl' in bdsim_properties.keys() or 'ksl' in bdsim_properties.keys(),
                **bdsim_properties)

        return bdsim_component

    def _build_bdsim_component(self, element: _pd.Series):
        bdsim_properties = self._build_bdsim_component_properties(element)

        if 'sequence' in self.model.keys():
            bdsim_element_type, properties = self._get_model_properties_for_element(element['NAME'])
            bdsim_properties = _merge(bdsim_properties, properties)
            element['PARENT'] = bdsim_element_type or BDSIM_RESERVED_NAME.get(element['PARENT'], element['PARENT'])
        else:
            element['PARENT'] = BDSIM_RESERVED_NAME.get(element['PARENT'], element['PARENT'])

        return self._generate_bdsim_component(element, bdsim_properties)

    def __call__(self,
                 with_bdsim_beam: bool = True,
                 with_bdsim_options: bool = True,
                 with_bdsim_placements: bool = True,
                 ):
        bdsim_input = _Builder.Machine()
        components = self.components[~self.components['BDSIM'].isnull()]
        # Add all components
        for name, component in components.iterrows():
            bdsim_input.Append(component['BDSIM'], is_component=True)

        # Add elements (won't be redefined, simply added to the sequence)
        for name, component in components.query("IS_ELEMENT == True").sort_values(by=['ID']).iterrows():
            bdsim_input.Append(component['BDSIM'], is_component=False)
            if 'apertureType' not in component['BDSIM'].keys() and 'xsize' not in component['BDSIM'].keys() \
                    and 'ysize' not in component['BDSIM'].keys():
                self.logging.warning(_log_component_info(name, component['PARENT'], component['BASE_PARENT'])
                                     + f" - Placing an element with no aperture definition !")

        if with_bdsim_beam:
            if 'twiss' not in self.madx.table:
                self.madx.input('twiss;')
            tw = self.madx.table['twiss'].dframe().iloc[0]
            bdsim_beam = _Beam.Beam(particletype=self.madx_beam['particle'],
                                           energy=self.madx_beam['energy'],
                                           distrtype='gausstwiss')
            bdsim_beam.SetX0(tw['x'])
            bdsim_beam.SetXP0(tw['px'])
            bdsim_beam.SetY0(tw['y'])
            bdsim_beam.SetYP0(tw['py'])
            bdsim_beam.SetBetaX(tw['betx'])
            bdsim_beam.SetBetaY(tw['bety'])
            bdsim_beam.SetAlphaX(tw['alfx'])
            bdsim_beam.SetAlphaY(tw['alfy'])
            bdsim_beam.SetDispX(tw['dx'])
            bdsim_beam.SetDispY(tw['dy'])

            bdsim_beam.SetEmittanceX(self.madx_beam['ex'])
            bdsim_beam.SetEmittanceY(self.madx_beam['ey'])
            bdsim_beam.SetSigmaE(self.madx_beam['sige'])
            #bdsim_beam.SetSigmaT(self.madx_beam['sigt'])
            bdsim_input.AddBeam(bdsim_beam)

        if with_bdsim_options:
            bdsim_input.AddOptions(_Options.Options(**self.model['options']))

        # if with_bdsim_placements:
        #     for ref_name, placements in self.placement_properties.items():
        #         for plcm in placements:
        #             for plm_name, placement_prop in plcm.items():
        #                 placement_prop['referenceElement'] = ref_name
        #                 bdsim_input.AddPlacement(plm_name + '_' + ref_name, **placement_prop)

        # if with_blms:
        #     pass
        #     # Add BLM
        #     for ref_name, blms in self.blms_properties.items():
        #         for blm in blms:
        #             for blm_name, blm_prop in blm.items():
        #                 blm_prop['referenceElement'] = ref_name
        #                 bdsim_input.AddBLM(blm_name + '_' + ref_name, **blm_prop)
        #

        #
        # if with_individual_placement:
        #     for ref_name, placements in self.aperture_file['placements'].items():
        #         bdsim_input.AddPlacement(ref_name, **placements)
        #

        return bdsim_input



"""
Writer

Write files for a pybdsim.Builder.Machine instance. Each section of the written output
(e.g. components, sequence, beam etc.) can be written in the main gmad file, written
in its own separate file, or called from an external, pre-existing file.

Classes:
File - A class that represents each section of the written output - contains booleans and strings.
Writer - A class that writes the data to disk.

"""
from pybdsim import __version__ as _pybdsimVersion
from . import _General
from . import Beam as _Beam
from . import Options as _Options

import time as _time
import os as _os
import textwrap as _textwrap
sections = ['components',
            'sequence',
            'samplers',
            'beam',
            'options',
            'bias',
            'material',
            'objects']

class FileSection():
    """
    A class that represents a section of a gmad file. The sections that this
    class can represent are:

    * Components
    * Sequence
    * Samplers
    * Beam
    * Options
    * Bias
    * Material


    The class contains booleans and strings relating to the location of
    that sections data. The section can set to be:

    * Written in its own separate file (default)
    * Written in the main gmad file
    * Called from an external file

    These classes are instantiated in the writer class for each section.
    An optional string passed in upon class instantiation is purely for
    the representation of the object which will state where the data will
    be written/called. This string should be one of the section names
    listed above.

    Example:

    >>> beam = FileSection('beam')
    >>> beam.CallExternalFile('../myBeam.gmad')
    >>> beam
    pybdsim.Writer.File instance
    File data will be called from the external file:
    ../myBeam.gmad

    """

    def __init__(self,willContain=''):
        #bool for file location/calling
        self._writeInMain         = False
        self._isWrittenSeparately = True
        self._isUserDefined       = False
        self._filePath            = ''

        #bool for written status, used to determine if required sections have been written
        self._hasBeenWritten      = False

        #string for __repr__ output
        self._willContain=''
        if isinstance(willContain, str) and willContain in sections:
            self._willContain = willContain

    def __repr__(self):
        s = ''
        s += 'pybdsim.Writer.File instance\n'

        if self._writeInMain:
            s += 'File data will be written into the main file.\n'
        elif self._isWrittenSeparately:
            if self._willContain != '':
                self._willContain += ' '
            s += 'File data will be written into a separate '+self._willContain+'file.\n'
        elif self._isUserDefined:
            s += 'File data will be called from the external file:\n'
            s += (self._filePath+'\n')
        return s

    def CallExternalFile(self, filepath=''):
        if not isinstance(filepath, str):
            raise TypeError("Filepath must be a string.")
        self._isUserDefined       = True
        self._isWrittenSeparately = False
        self._writeInMain         = False
        self._filePath            = filepath
        self._hasBeenWritten      = True

    def WriteInMain(self):
        self._isUserDefined       = False
        self._isWrittenSeparately = False
        self._writeInMain         = True
        self._filePath            = ''

    def WriteSeparately(self):
        self._isUserDefined       = False
        self._isWrittenSeparately = True
        self._writeInMain         = False
        self._filePath            = ''


class Writer():
    """
    A class for writing a pybdsim.Builder.Machine instance to file.

    This class allows the user to write individual sections of a BDSIM input file
    (e.g. components, sequence, beam etc.) or write the machine as a whole.

    There are 6 attributes in this class which are FileSection instances representing each
    section of the data. The location where these sections will be written/read is
    stored in these instances. See the FileSection class for further details.

    The optional boolean 'singlefile' in the WriteMachine function for writing
    the sections to a single file overrides any sections locations set in their
    respective FileSection instances.

    This class also has individual functions (e.g. WriteBeam) to write each file section
    and the main file (WriteMain) separately. These section functions must be called BEFORE
    the WriteMain function is called otherwise the main file will have no reference to
    these sections.

    Examples:

    Writing the Builder.Machine instance myMachine to separate files:

    >>> a = Writer()
    >>> a.WriteMachine(myMachine,'lattice.gmad')
    Lattice written to:
    lattice_components.gmad
    lattice_sequence.gmad
    lattice_beam.gmad
    lattice.gmad
    All included in main file:
    lattice.gmad

    Writing the Builder.Machine instance myMachine into a single file:

    >>> a = Writer()
    >>> a.WriteMachine(myMachine,'lattice.gmad',singlefile=True)
    Lattice written to:
    lattice.gmad
    All included in main file:
    lattice.gmad

    """
    def __init__(self):
        #FileSection instances for each section of the input.
        self.Components = FileSection('components')
        self.Sequence   = FileSection('sequence')
        self.Samplers   = FileSection('samplers')
        self.Beam       = FileSection('beam')
        self.Options    = FileSection('options')
        self.Objects    = FileSection('options')
        self.Bias       = FileSection('bias')
        self.Material   = FileSection('material')

        self._mainFileLines   = []        #lines that will be written to the main file.
        self._elementsperline = 100       #number of machine elements per bdsim line (not text line)
        self._basefilename    = 'lattice' #default base file name
        self._timestring      = '! ' + _time.strftime("%a, %d %b %Y %H:%M:%S +0000", _time.gmtime()) + '\n'

        #default filenames. These can be overwritten to a user defined filename.
        self._mainFilename = self._basefilename + '.gmad'
        self._defaultSectionFilenames  = {}
        for sectiontype in sections:
            self._defaultSectionFilenames[sectiontype] = self._basefilename + '_' + sectiontype + '.gmad'

        #list of sections that will be / have been written. This is for both user feedback and
        #for any include lines that will be written in the main file.
        self._sectionsToBeWritten = []

    def WriteMachine(self,machine, filename, singlefile=False,
                     verbose=True, overwrite=True):
        """
        WriteMachine(machine(machine),filename(string),singlefile(bool),verbose(bool))

        Write a machine to disk. By default, the machine will be written
        into the following individual files:

        +---------------------------+----------------------------------------+
        | filename_components.gmad  | component files (max 10k per file)     |
        +---------------------------+----------------------------------------+
        | filename_sequence.gmad    | lattice definition                     |
        +---------------------------+----------------------------------------+
        | filename_samplers.gmad    | sampler definitions (max 10k per file) |
        +---------------------------+----------------------------------------+
        | filename_options.gmad     | options                                |
        +---------------------------+----------------------------------------+
        | filename_beam.gmad        | beam definition                        |
        +---------------------------+----------------------------------------+
        | filename_bias.gmad        | machine biases (if defined)            |
        +---------------------------+----------------------------------------+
        | filename_material.gmad    | machine materials (if defined)         |
        +---------------------------+----------------------------------------+
        | filename.gmad             | suitable main file with all sub        |
        |                           | files in correct order                 |
        +---------------------------+----------------------------------------+

        These are prefixed with the specified filename / path

        The optional bool singlefile = True will write all the above sections
        into a single file:

        filename.gmad

        kwargs:
        overwrite : Do not append an integer to the basefilename if
        already exists, instead overwrite existing files.

        """
        self._checkFiles(filename, overwrite)

        if singlefile:
            #set all sections to write in the main file
            self.Components.WriteInMain()
            self.Sequence.WriteInMain()
            self.Beam.WriteInMain()
            self.Options.WriteInMain()
            self.Samplers.WriteInMain()
            self.Bias.WriteInMain()
            self.Material.WriteInMain()
            if len(machine.objects) > 0:
                self.Objects.WriteInMain()

        #write the individual files.
        self.WriteComponents(machine)
        self.WriteSequence(machine)
        self.WriteSamplers(machine)
        self.WriteBeam(machine)
        self.WriteOptions(machine)
        self.WriteBias(machine)
        self.WriteMaterial(machine)
        if len(machine.objects) > 0:
            self.WriteObjects(machine)

        #Write main
        self.WriteMain(machine)

        if verbose:
            #user feedback
            print('Lattice written to:')
            for section in self._sectionsToBeWritten:
                sectObject = getattr(self,section) #a copy of the FileSection object.
                fn = getattr(sectObject,'_filePath')
                print(fn)
            print('All included in main file: \n',self._mainFilename)

    def WriteMain(self,machine,filename=''):
        """
        WriteMain(machine(machine),filename(string))

        Write the main gmad file:
        filename.gmad

        The functions for the other sections of the machine (components,sequence,beam,options,samplers,bias,material)
        must be written BEFORE this function is called.

        """

        if filename == '':
            fn_main = self._mainFilename #default
        else:
            fn_main = filename #override

        #do not write if the beam, components or sequence sections have not been written.
        #These three have to exist for a machine to work.
        if (not self.Components._hasBeenWritten) or (not self.Sequence._hasBeenWritten) or (not self.Beam._hasBeenWritten):
            exceptionString = "The following sections must be written before the main file can be written:\n"
            exceptionString += "- Components\n- Sequence\n- Beam"
            raise Exception(exceptionString)

        #write main file
        with open(fn_main, 'w') as f:
            self._writeFileheader(
                f, ["! Number of Elements:\t{}".format(len(machine.elements)),
                    "! Total length:\t\t{}m".format(machine.length)])

            # pre includes
            if hasattr(machine, "includesPre"):
                for inc in machine.includesPre:
                    f.write('include ' + inc + ";\n")

            #other files to include
            for section in self._sectionsToBeWritten:
                sectObject = getattr(self,section)
                fn = getattr(sectObject,'_filePath')
                isUserDefined = getattr(sectObject,'_isUserDefined')
                if not isUserDefined:
                    fn = fn.split('/')[-1]
                f.write('include '+fn+';\n')
            f.write('\n\n')

            # write lines to main file from components, sequence etc.
            for line in self._mainFileLines:
                f.write(line)
            #write samplers in this main file if less than 10 samplers.
            if len(machine.samplers) <= 10:
                for sampler in machine.samplers:
                    f.write(str(sampler))

            # post includes
            if hasattr(machine, "includesPost"):
                for inc in machine.includesPost:
                    f.write('include ' + inc + ";\n")
        self._mainFilename = fn_main

    def WriteComponents(self,machine,filename=''):
        """
        Write the machines components to disk:
        filename.gmad
        """
        fn_components = self._getName(filename,'components')
        if self.Components._writeInMain:                #if _writeInMain, append strings to _mainFileLines list.
            for element in list(machine.elements.values()):
                self._mainFileLines.append(str(element))
            self._mainFileLines.append('\r\n')
        elif self.Components._isWrittenSeparately:      #if _isWrittenSeparately, write directly to file here.
            with open(fn_components, 'w') as f:
                self._writeFileheader(f, ['! COMPONENT DEFINITION'])
                for element in list(machine.elements.values()):
                    f.write(str(element))
            self._sectionsToBeWritten.append('Components')
            self.Components._filePath = fn_components   #update FileSection path
        elif self.Components._isUserDefined:
            self._sectionsToBeWritten.append('Components')
        self.Components._hasBeenWritten = True

    def WriteBias(self,machine,filename=''):
        """
        Write the machines bias to disk:
        filename.gmad
        """
        fn_bias = self._getName(filename,'bias')

        #write bias if it exists
        if len(machine.bias) > 0:
            if self.Bias._writeInMain:
                for bias in machine.bias:
                    self._mainFileLines.append(str(bias))
                self._mainFileLines.append('\r\n')
            elif self.Bias._isWrittenSeparately:
                with open(fn_bias, 'w') as f:
                    self._writeFileheader(f, ["! BIAS DEFINITION"])
                    for bias in machine.bias:
                        f.write(str(bias))
                self.Bias._filePath = fn_bias
                self._sectionsToBeWritten.append('Bias')
        elif self.Bias._isUserDefined:
            self._sectionsToBeWritten.append('Bias')
        self.Bias._hasBeenWritten = True

    def WriteMaterial(self,machine,filename=''):
        """
        Write the machines material to disk:
        filename.gmad
        """
        fn_material = self._getName(filename,'material')

        #write matrial if it exists
        if len(machine.material) > 0:
            if self.Material._writeInMain:
                for material in machine.material:
                    self._mainFileLines.append(str(material))
                self._mainFileLines.append('\r\n')
            elif self.Material._isWrittenSeparately:
                with open(fn_material, 'w') as f:
                    self._writeFileheader(f, ["! MATERIAL DEFINITION"])
                    for material in machine.material:
                        f.write(str(material))
                self.Material._filePath = fn_material
                self._sectionsToBeWritten.append('Material')
        elif self.Material._isUserDefined:
            self._sectionsToBeWritten.append('Material')
        self.Material._hasBeenWritten = True

    def WriteBeam(self,machine,filename=''):
        """
        Write a machines beam to disk:
        filename.gmad

        Machine can be either a pybdsim.Builder.Machine instance
        or a pybdsim.Beam.Beam instance.
        """
        #check data type
        if isinstance(machine,_Beam.Beam):
            object = machine
        else:
            object = machine.beam

        fn_beam = self._getName(filename,'beam')

        if self.Beam._writeInMain:
            self._mainFileLines.append(object.ReturnBeamString())
            self._mainFileLines.append('\r\n')
        elif self.Beam._isWrittenSeparately:
            with open(fn_beam, 'w') as f:
                self._writeFileheader(f, ["! BEAM DEFINITION"])
                f.write(object.ReturnBeamString())
            self.Beam._filePath = fn_beam
            self._sectionsToBeWritten.append('Beam')
        elif self.Beam._isUserDefined:
            self._sectionsToBeWritten.append('Beam')
        self.Beam._hasBeenWritten = True

    def WriteSamplers(self,machine,filename=''):
        """
        Write the machines samplers to disk:
        filename.gmad
        """
        fn_samplers = self._getName(filename,'samplers')

        #write samplers
        # if less than 10 samplers, just put in main file
        if len(machine.samplers) > 10:
            if self.Samplers._writeInMain:
                for sampler in machine.samplers:
                    self._mainFileLines.append(str(sampler))
                self._mainFileLines.append('\r\n')
            elif self.Samplers._isWrittenSeparately:
                with open(fn_samplers, 'w') as f:
                    self._writeFileheader(f, ["! SAMPLER DEFINITION"])
                    for sampler in machine.samplers:
                        f.write(str(sampler))
                self.Samplers._filePath = fn_samplers
                self._sectionsToBeWritten.append('Samplers')
        elif self.Samplers._isUserDefined:
            self._sectionsToBeWritten.append('Samplers')
        self.Samplers._hasBeenWritten = True

    def WriteObjects(self,machine,filename=''):
        """
        Write the machines objects (e.g. crystals) to disk:
        filename.gmad
        """
        fn_objects = self._getName(filename,'objects')

        if self.Objects._writeInMain:
            for obj in machine.objects:
                self._mainFileLines.append(str(obj))
            self._mainFileLines.append('\r\n')
        else:
            with open(fn_objects, 'w') as f:
                self._writeFileheader(f, ["! OBJECTS DEFINITION"])
                for obj in machine.objects:
                    f.write(str(obj))
            self.Objects._filePath = fn_objects
            self._sectionsToBeWritten.append('Objects')
        self.Objects._hasBeenWritten = True

    def WriteOptions(self,machine,filename=''):
        """
        Write a machines options to disk:
        filename.gmad

        Machine can be either a pybdsim.Builder.Machine instance
        or a pybdsim.Options.Options instance.
        """
        #check data type
        if isinstance(machine,_Options.Options):
            object = machine
        else:
            object = machine.options

        fn_options = self._getName(filename,'options')

        # write options - only if specified
        if object != None:
            if self.Options._writeInMain:
                self._mainFileLines.append(object.ReturnOptionsString())
                self._mainFileLines.append('\r\n')
            elif self.Options._isWrittenSeparately:
                with open(fn_options, 'w') as f:
                    self._writeFileheader(f, ['! OPTIONS DEFINITION'])
                    f.write(object.ReturnOptionsString())
                self.Options._filePath = fn_options
                self._sectionsToBeWritten.append('Options')
        elif self.Options._isUserDefined:
            self._sectionsToBeWritten.append('Options')
        self.Options._hasBeenWritten = True

    def WriteSequence(self,machine,filename=''):
        """
        Write the machines sequence to disk:
        filename.gmad
        """
        fn_sequence = self._getName(filename,'sequence')

        # need to define the period before making sampler planes
        sequencechunks = _General.Chunks(machine.sequence,self._elementsperline)
        linelist = []
        ti = 0

        #write lattice sequence
        if self.Sequence._writeInMain:
            for line in _General.Chunks(machine.sequence,self._elementsperline):
                # Use _textwrap.wrap to wrap very long lines
                linetxt = '\n\t'.join(_textwrap.wrap(
                    "l{}: line = ({});".format(ti, ', '.join(line))))
                self._mainFileLines.append("{}\n".format(linetxt))
                linelist.append('l'+str(ti))
                ti += 1
            self._mainFileLines.append('lattice: line = ('+', '.join(linelist)+');\n')
            self._mainFileLines.append('use, period=lattice;\n')
            self._mainFileLines.append('\r\n')
        elif self.Sequence._isWrittenSeparately:
            with open(fn_sequence, 'w') as f:
                self._writeFileheader(f, ["! LATTICE SEQUENCE DEFINITION"])
                for line in _General.Chunks(machine.sequence,self._elementsperline):
                    # Use _textwrap.wrap to wrap very long lines
                    linetxt = '\n\t'.join(_textwrap.wrap(
                        "l{}: line = ({});".format(ti, ', '.join(line))))
                    f.write("{}\n".format(linetxt))
                    linelist.append('l'+str(ti))
                    ti += 1

                f.write('lattice: line = ('+', '.join(linelist)+');\n')
                f.write('use, period=lattice;\n')
            self.Sequence._filePath = fn_sequence
            self._sectionsToBeWritten.append('Sequence')
        elif self.Sequence._isUserDefined:
            self._sectionsToBeWritten.append('Sequence')
        self.Sequence._hasBeenWritten = True

    def _getName(self,filename,sectiontype=''):
        #check input types are strings
        if not isinstance(filename, str):
            raise TypeError("Filename not a string")
        if not isinstance(sectiontype, str):
            raise TypeError("Sectiontype not a string")

        if filename == '' and sectiontype == '':
            raise ValueError("Both filename and sectiontype cannot be empty strings")

        #get section file name
        if sectiontype in sections:
            if (filename == '') or (filename == self._basefilename):
                fn_name = self._defaultSectionFilenames[sectiontype]    #no name or name = basefilename, use section default
            else:
                fn_name = self._checkExtensionAndPath(filename)         #otherwise, override the filename
        else:
            if (filename == '') or (filename == self._basefilename):    #no section but filename clashes with basefilename
                errorString =  "Unknown section type - filename cannot be an empty string "
                errorString += "or match the basefilename ("+str(self._basefilename)+")."
                raise ValueError(errorString)
            else:
                fn_name = self._checkExtensionAndPath(filename)         #override filename

        #check for duplicate file names.
        for section in self._sectionsToBeWritten:
            sectObject = getattr(self,section)
            fn = getattr(sectObject,'_filePath')
            if (fn == fn_name) or (fn == self._mainFilename):
                raise ValueError("Filename already used.")
        return fn_name

    def _checkFiles(self,filename, overwrite=True):
        filename = self._checkExtensionAndPath(filename)

        if not overwrite:
            #check if file already exists
            originalFilename = filename
            filename = _General.GenUniqueFilename(filename)
            if filename != originalFilename:
                print('Warning, chosen filename already exists - using filename: ',filename.split('.')[0])

        basefilename = filename[:-5] #everything before '.gmad'
        #new default section names
        self._defaultSectionFilenames['components'] = basefilename + '_components.gmad'
        self._defaultSectionFilenames['sequence']   = basefilename + '_sequence.gmad'
        self._defaultSectionFilenames['samplers']   = basefilename + '_samplers.gmad'
        self._defaultSectionFilenames['beam']       = basefilename + '_beam.gmad'
        self._defaultSectionFilenames['options']    = basefilename + '_options.gmad'
        self._defaultSectionFilenames['bias']       = basefilename + '_bias.gmad'
        self._defaultSectionFilenames['material']   = basefilename + '_material.gmad'
        self._defaultSectionFilenames['objects']    = basefilename + '_objects.gmad'
        self._mainFilename = basefilename + '.gmad'
        self._basefilename = basefilename
        self._timestring = '! ' + _time.strftime("%a, %d %b %Y %H:%M:%S +0000", _time.gmtime()) + '\n'

    def _checkExtensionAndPath(self,filename):
        #check filename for extension
        if filename[-5:] != '.gmad':
            filename += '.gmad'

        #check for directory and make it if not:
        if '/' in filename:
            directory = '/'.join(filename.split('/')[:-1]) #strip the filename off
            if not _os.path.exists(directory):
                _os.system("mkdir -p " + directory)
        return filename

    def _writeFileheader(self, f, extralines):
        """Write the comment header for a gmad file.  Format:
        date and time
        pybdsim version info
        extraline1
        extraline2
        ...


        the gmad file...

        extra lines should be provided without newlines chars at the end.
        """
        f.write(self._timestring)
        f.write(
            '! pybdsim.Builder Lattice, using pybdsim version {} \n'.format(
                _pybdsimVersion))
        for extraline in extralines:
            f.write("{}\n".format(extraline))
        if extralines:
            f.write('\n')
        else:
            f.write('\n\n')



import pandas as _pd
import numpy as _np
import matplotlib.pyplot as plt
from cpymad.madx import Madx

import pymadx
import pybdsim
from pybdsim.DataUproot import BDSIMOutput
from pybdsim.Optics import Twiss
import georges_core
from georges_core.units import ureg as _ureg


if __name__ == "__main__":

    madx_input = f"""
        BEAM, PARTICLE=PROTON, PC = 2.794987;
        BRHO      := BEAM->PC * 3.3356;

        call, file="madx/ps_mu.seq";
        call, file="madx/ps_ss.seq";
        call, file="madx/scenarios/{name_path}/{sub_scenario}/{scenario}.beam";
        call, file="madx/scenarios/{name_path}/{sub_scenario}/{scenario}.str";

        use, sequence=PS;
        select,flag=twiss, clear;
        twiss, sequence=PS, file='madx_optics/{name_path}.tfs';
        
        stop;
    """
    madx = Madx()
    madx.input(madx_input)
    tw = madx.table['twiss'].dframe()

    ps_machine = pybdsim.Convert.CPyMad2Gmad(madx_instance=madx,
                                             model_file=f"model/{model_type}.yml")(with_bdsim_beam=False)
    madx_beam = madx.sequence['ps'].beam
    offset = 1e-5
    x0 = tw.iloc[0]['x']
    y0 = tw.iloc[0]['y']
    xp0 = tw.iloc[0]['px']
    yp0 = tw.iloc[0]['py']

    twiss_beam = _np.array([[x0, xp0, y0, yp0, madx_beam['pc']],
                            [x0 + offset, xp0, y0, yp0, madx_beam['pc']],
                            [x0, xp0 + offset, y0, yp0, madx_beam['pc']],
                            [x0, xp0, y0 + offset, yp0, madx_beam['pc']],
                            [x0, xp0, y0, yp0 + offset, madx_beam['pc']],
                            [x0, xp0, y0, yp0, (1 + offset) * madx_beam['pc']],
                            [x0 - offset, xp0, y0, yp0, madx_beam['pc']],
                            [x0, xp0 - offset, y0, yp0, madx_beam['pc']],
                            [x0, xp0, y0 - offset, yp0, madx_beam['pc']],
                            [x0, xp0, y0, yp0 - offset, madx_beam['pc']],
                            [x0, xp0, y0, yp0, (1 - offset) * madx_beam['pc']]
                            ])

    _np.savetxt(f'{path}twiss_coordinates.dat', twiss_beam, delimiter=' ', fmt='%.8e')

    bdsim_beam = pybdsim.Beam.Beam(particletype=madx_beam['particle'],
                                   energy=madx_beam['energy'],
                                   distrtype="userfile",
                                   distrFile="\"twiss_coordinates.dat\"",
                                   distrFileFormat="\"x[m]:xp[rad]:y[m]:yp[rad]:P[GeV]\"")
    ps_machine.AddBeam(bdsim_beam)
    ps_machine.options['samplersSplitLevel'] = 0
    ps_machine.Write(filename=f'{path}ps.gmad')

    # Add sample all at the end of the main file
    with open(f"{path}ps.gmad", 'a') as f:
        f.write('sample, all;')

    # opening the file in read mode
    file = open(f"{path}ps_beam.gmad", "r")
    replacement = ""
    # using the for loop
    for line in file:
        line = line.strip()
        if 'energy' in line:
            changes = f"momentum = {madx_beam['pc']}*GeV,"
        else:
            changes = line
        replacement = replacement + changes + "\n"

    file.close()
    # opening the file in write mode
    fout = open(f"{path}ps_beam.gmad", "w")
    fout.write(replacement)
    fout.close()

    pybdsim.Run.Bdsim(gmadpath=f"{path}ps.gmad",
                      outfile=f"{path}results/output_{scenario}_{model_type}",
                      ngenerate=11,
                      batch=True,
                      bdsimExecutable="/Users/rtesse/packages/bdsim-build-clion/install/bin/bdsim")

    samplers = pybdsim.Data.Load(filepath=f"{path}results/output_{scenario}_{model_type}.root")
    sampler_name = list(map(str, list(samplers.GetSamplerNames())))[0:-2]  # last element has no data if circular
    sampler_name[0] = 'drift_0.'
    samplers_data = {}

    for idx, sm in enumerate(sampler_name):
        # Careful, order matters
        data_samplers = _pd.DataFrame(data=dict((k,
                                                 pybdsim.Data.SamplerData(samplers, sm).data[k]) for k in
                                                ('x', 'xp', 'y', 'yp', 'S', 'p')))

        samplers_data[sm] = data_samplers

    tw_init = {'ALPHA11': tw.iloc[0]['alfx'], 'ALPHA22': tw.iloc[0]['alfy'],
               'BETA11': tw.iloc[0]['betx'], 'BETA22': tw.iloc[0]['bety'],
               'DISP1': tw.iloc[0]['dx'] * beta_rel, 'DISP2': tw.iloc[0]['dpx'] * beta_rel,
               'DISP3': tw.iloc[0]['dy'] * beta_rel, 'DISP4': tw.iloc[0]['dpy'] * beta_rel
               }

    tw_init['GAMMA11'] = (1 + tw.iloc[0]['alfx'] ** 2) / tw.iloc[0]['betx']
    tw_init['GAMMA22'] = (1 + tw.iloc[0]['alfy'] ** 2) / tw.iloc[0]['bety']


#!/usr/bin/env python2.7

import pickle
import ROOT as _ROOT 
from ROOT import TFile as _TFile 
import root_numpy as _root_numpy 
import sys as _sys

def Rebdsim2Numpy(rootfilename,picklefilename):
    """
    Load an output file from rebdsim using root_numpy, convert
    to a dictionary of arrays and save in a Python binary pickled file.
    
    Arguments Rebdsim2Numpy(inputfilename, outputfilename)
    """
    # open file 
    f = _TFile(rootfilename)

    # loop over keys
    kl = f.GetListOfKeys() 

    # data dict
    dd = {}

    for ki in range(0,kl.GetSize()) :
        cname = kl[ki].GetClassName()
        name  = kl[ki].GetName()

        if cname == "TDirectoryFile" : 
            d = f.Get(name)
            d.cd()
            # loop over histos in dir
            dkl = d.GetListOfKeys()
            for dki in range(0,dkl.GetSize()) :
                h  = d.Get(dkl[dki].GetName())
                ah = _root_numpy.hist2array(h)
                print("Histogram : ",name+"_"+dkl[dki].GetName())
                dd[name+"_"+dkl[dki].GetName()] = ah

        if cname == "TTree" : 
            t = f.Get(name)
            print("Tree      : ",name)
            at = _root_numpy.tree2array(t)
            dd[name] = at

     
    f.Close()
   
    of = open(picklefilename,"wb")
    pickle.dump(dd,of)
    of.close()

if __name__ == "__main__":
    nargs = len(_sys.argv)
    if (nargs < 3 or nargs > 3):
        print("Error - Usage: Rebdsim2Numpy.py inputfile outputfile")
        _sys.exit(1)
    Rebdsim2Numpy(_sys.argv[1],_sys.argv[2])



from .. import Builder as _Builder
from .. import Beam as _Beam
from  pymad8.Output import Saveline as _Saveline

# pybdsim.Convert.Mad8Saveline2Gmad('../ebds.saveline', 'ilc.gmad', ignore_zero_length_items=False)

def Mad8Saveline2Gmad(input, output_file_name, start_name=None, end_name=None, ignore_zero_length_items=True,
                      samplers='all', aperture_dict={}, collimator_dict="collimators.dat", beam_pipe_radius=0.2,
                      verbose=False, beam=True, optics=True, loss=True):
    items_omitted = []  # Store any items that have been skipped over.
    fake_length = 1e-6

    if type(input) == str:
        mad8 = _Saveline.Loader(input)
    else:
        mad8 = input

    if type(collimator_dict) == str :
        collimator_obj = Mad8SavelineCollimatorDatabase(collimator_dict)
        collimator_dict = collimator_obj.GetDict()
        xsize_dict = 'xsize'
        ysize_dict = 'ysize'
    else:  # Necessary as the dictionaries use different cases.
        xsize_dict = 'XSIZE'
        ysize_dict = 'YSIZE'
    
    ilc = _Builder.Machine()

    for element in mad8.elementList:
        element_type = list(mad8.elementDict[element].keys())[0]
        element_properties = list(mad8.elementDict[element].values())[0]
        length = element_properties['L'] if 'L' in element_properties else fake_length  # Everything has a length.
        zero_length = True if length < 1e-9 else False

        if element_type == 'LINE':
            for name in element_properties:
                sequence_element_type = list(mad8.elementDict[name].keys())[0]
                sequence_element_properties = list(mad8.elementDict[name].values())[0]
                # Everything has a length.
                length = sequence_element_properties['L'] if 'L' in sequence_element_properties else fake_length

                construct_sequence(ilc, name, sequence_element_type, sequence_element_properties, length,
                                   ignore_zero_length_items, items_omitted, zero_length, aperture_dict, collimator_dict,
                                   beam_pipe_radius, xsize_dict, ysize_dict, verbose)

    ilc.AddSampler(samplers)

    if beam:  # Use ILC defaults
        beam = _Beam(particletype='e-', energy=250, distrtype='gausstwiss')

        beam._SetBetaX(71.482996720846, unitsstring='m')
        beam._SetBetaY(39.603963961143, unitsstring='m')
        beam._SetAlphaX(-1.562625067837)
        beam._SetAlphaY(1.283201237991)
        beam.SetZ0(-1, unitsstring='um')
        beam.SetT0(1e-15)
        beam._SetEmittanceX(emitx=2.044e-11,unitsstring='m')
        beam._SetEmittanceY(emity=8.176e-14, unitsstring='m')

        ilc.AddBeam(beam)

    if optics:
        ilc.options.SetPhysicsList('QGSP_BERT')

    if loss:
        ilc.options.SetStopTracks(False)
            
    ilc.options.SetNGenerate(1000)
    ilc.Write(output_file_name)


def construct_sequence(ilc, element, element_type, element_properties, length, ignore_zero_length_items,
                       items_omitted, zero_length, aperture_dict, collimator_dict, beam_pipe_radius, xsize_dict, ysize_dict, verbose):
    kws = {}
    if element_type == 'DRIFT':
            ilc.AddDrift(element, length, **kws)

    elif element_type == 'VKICKER':
        kws['kick'] = element_properties['KICK'] if 'KICK' in element_properties else 0

        ilc.AddVKicker(element, length, **kws)

    elif element_type == 'HKICKER':
        kws['kick'] = element_properties['KICK'] if 'KICK' in element_properties else 0

        ilc.AddHKicker(element, length, **kws)

    elif element_type == 'MARKER':
        if ignore_zero_length_items:
            items_omitted.append(element)
        else:
            ilc.AddMarker(element)

    elif element_type == 'SBEND':
        angle = element_properties['ANGLE'] if 'ANGLE' in element_properties else 0
        k1 = (element_properties['K1'] / length) if 'K1' in element_properties else 0

        kws['hgap'] = element_properties['HGAP'] if 'HGAP' in element_properties else 0
        kws['fintx'] = element_properties['FINTX'] if 'FINTX' in element_properties else 0
        kws['tilt'] = element_properties['TILT'] if 'TILT' in element_properties else 0
        kws['fint'] = element_properties['FINT'] if 'FINT' in element_properties else 0
        kws['e1'] = element_properties['E1'] if 'E1' in element_properties else 0
        kws['e2'] = element_properties['E2'] if 'E2' in element_properties else 0

        ilc.AddDipole(element, 'sbend', length, angle, k1=k1, **kws)

    elif element_type == 'RBEND':
        angle = element_properties['ANGLE'] if 'ANGLE' in element_properties else 0

        ilc.AddDipole(element, 'rbend', angle, **kws)

    elif element_type == 'QUADRUPOLE':
        k1 = element_properties['K1'] if 'K1' in element_properties else 0

        #kws['aperture'] = element_properties['APERTURE'] if 'APERTURE' in element_properties else 0

        ilc.AddQuadrupole(element, length, k1, **kws)

    elif element_type == 'SEXTUPOLE':
        k2 = element_properties['K2'] if 'K2' in element_properties else 0

        kws['tilt'] = element_properties['TILT'] if 'TILT' in element_properties else 0

        #  Aperture not currently supported
        #kws['aperture'] = element_properties['APERTURE'] if 'APERTURE' in element_properties else 0 aperX / aperY

        #ilc.AddSextupole(element, length, k2, **kws)
        ilc.AddDrift(element,length)
    elif element_type == 'OCTUPOLE':
        k3 = element_properties['K3'] if 'K3' in element_properties else 0

        #  Aperture not currently supported.
        #kws['aperture'] = element_properties['APERTURE'] if 'APERTURE' in element_properties else 0

        # ilc.AddOctupole(element, length, k3, **kws)
        ilc.AddDrift(element, length)

    elif element_type == 'MULTIPOLE':  # Not fully implemented, but ready to go when AddMultipole works
        tilt = element_properties['TILT'] if 'TILT' in element_properties else 0
        knl = ()
        kns = ()

        for i in range(1, 7):  # Range is fixed to between 1 - 6 as far as I can tell, might be more though
            key_l = 'K%sL' % i
            key_s = 'K%sS' % i
            temp = element_properties[key_l] if key_l in element_properties else 0
            temp2 = element_properties[key_s] if key_s in element_properties else 0
            knl = knl + (temp,)
            kns = kns + (temp2,)

        kws['lrad'] = element_properties['LRAD'] if 'LRAD' in element_properties else 0
        #  Aperture not currently supported
        #kws['aperture'] = element_properties['APERTURE'] if 'APERTURE' in element_properties else 0

        # ilc.AddMultipole(element, length, knl, ksl=kns, tilt=tilt, **kws)  # ksl should be kns to keep format
        ilc.AddMarker(element)

    elif element_type == 'ECOLLIMATOR': # Need to check logic precedence here with Boog.
        if element in collimator_dict:
            kws['material'] = collimator_dict[element]['bdsim_material'] if 'bdsim_material' in collimator_dict else 'Copper'
            xsize = collimator_dict[element][xsize_dict] if xsize_dict in collimator_dict[element] else beam_pipe_radius
            ysize = collimator_dict[element][ysize_dict] if ysize_dict in collimator_dict[element] else beam_pipe_radius
#            angle = collimator_dict[element]['ANGLE'] if 'ANGLE' in collimator_dict[element] else 0.0
        else:
            kws['material'] = 'Copper'
            xsize = element_properties[xsize_dict] if xsize_dict in element_properties else beam_pipe_radius
            ysize = element_properties[ysize_dict] if ysize_dict in element_properties else beam_pipe_radius
#            angle = element_properties['ANGLE'] if 'ANGLE' in element_properties else 0

        ilc.AddECol(element, length, xsize, ysize, **kws)

    elif element_type == 'RCOLLIMATOR':
        if element in collimator_dict:
            kws['material'] = collimator_dict[element]['bdsim_material'] if 'bdsim_material' in collimator_dict else 'Copper'
            xsize = collimator_dict[element][xsize_dict] if xsize_dict in collimator_dict[element] else beam_pipe_radius
            ysize = collimator_dict[element][ysize_dict] if ysize_dict in collimator_dict[element] else beam_pipe_radius
 #           angle = collimator_dict[element]['ANGLE'] if 'ANGLE' in collimator_dict[element] else 0.0
        else:
            kws['material'] = 'Copper'
            xsize = element_properties[xsize_dict] if xsize_dict in element_properties else beam_pipe_radius
            ysize = element_properties[ysize_dict] if ysize_dict in element_properties else beam_pipe_radius
#            angle = element_properties['ANGLE'] if 'ANGLE' in element_properties else 0

        ilc.AddRCol(element, length, xsize, ysize, **kws)

    elif element_type == 'WIRE':
        if ignore_zero_length_items and zero_length:
            items_omitted.append(element)
        elif (not ignore_zero_length_items) and zero_length:
            ilc.AddMarker(element)
            if verbose:
                print(element, ' -> marker instead of wire.')
        else:
            ilc.AddDrift(element, length, **kws)

    elif element_type == 'INSTRUMENT':
        if ignore_zero_length_items and zero_length:
            items_omitted.append(element)
        elif (not ignore_zero_length_items) and zero_length:
            ilc.AddMarker(element)
            if verbose:
                print(element, ' -> marker instead of instrument.')
        else:
            ilc.AddDrift(element, length, **kws)

    elif element_type == 'MONITOR':
        if ignore_zero_length_items and zero_length:
            items_omitted.append(element)
        elif (not ignore_zero_length_items) and zero_length:
            ilc.AddMarker(element)
            if verbose:
                print(element, ' -> marker instead of monitor.')
        else:
            ilc.AddDrift(element, length, **kws)

    elif element_type == 'LCAVITY':
        deltae = element_properties['DELTAE'] * 1000 if 'DELTAE' in element_properties else 0  # MeV
        gradient = deltae / length

        #kws['aperture'] = element_properties['APERTURE'] if 'APERTURE' in element_properties else 0
        kws['freq'] = element_properties['FREQ'] if 'FREQ' in element_properties else 0
        kws['phi0'] = element_properties['PHI0'] if 'PHI0' in element_properties else 0
        kws['eloss'] = element_properties['ELOSS'] if 'ELOSS' in element_properties else 0

        ilc.AddRFCavity(element, length, gradient, **kws)


class Mad8SavelineCollimatorDatabase: 
    '''
    Load collimator file into memory and functions to open and manipulate collimator system
    c = Mad8CollimatorDataBase(fileName)
    fileName = "collimator.dat"
    file format
    <name> <angle> <length> <x_size/2> <ysize/2> <material>
    <length> includes the tapers, so entire length along machine 
    '''

    def __init__(self,collimatorFileName) :
        self.collimatorFileName = collimatorFileName
        self.LoadCollimatorDb(self.collimatorFileName) 
        
    def LoadCollimatorDb(self,collimatorFileName) : 
        f = open(collimatorFileName) 

        inx = 0 

        self._coll = {}
        for l in f : 
            t = l.split()
            name     = t[0]
            angle    = float(t[1])
            length   = float(t[2])
            xsize    = float(t[3])
            ysize    = float(t[4]) 
            material = t[5]
            inx = inx + 1 

            d = {'index':inx, 'angle':angle, 'l':length, 'xsize':xsize,
                 'ysize':ysize, 'bdsim_material':material}

            self._coll[name] = d     
        
    def OpenCollimators(self,openHalfSizeX=0.1, openHalfSizeY=0.1) : 
        for c in list(self._coll.keys()) :
            self._coll[c]['xsize'] = openHalfSizeX
            self._coll[c]['ysize'] = openHalfSizeY
    
    def GetCollimators(self) : 
        return list(self._coll.keys())

    def GetCollimator(self, name) : 
        return self._coll[name]

    def GetDict(self) : 
        return self._coll



import pymadx.Ptc
import pymadx.Beam
import pymadx.Builder
import pymadx.Tfs
import pybdsim.Beam
import pybdsim.Builder
import pybdsim.Data
import os as _os
import matplotlib.pyplot as _plt
import robdsim
import numpy as _np

class Test:
    def __init__(self,type_,foldername=None,particle="e-",energy=1.0,distribution='flat',nparticles=10,length=1.0,**kwargs): 
        """
        Tracking test class

        type = 'drift' | 'quadrupole' | 'sextupole' | 'sbend' | 'solenoid'
        distribution = 'flat' | 'gaussian'
        """
        self.type_        = type_
        self.filename     = self.type_
        self.foldername   = foldername
        self.ptcfilename  = 'inrays.madx'
        if self.foldername != None:
            self.usingfolder = True
            self.filepath = self.foldername+'/'+self.filename
            self.ptcfilepath =  self.foldername+'/'+self.ptcfilename
            _os.system("mkdir -p " + self.foldername)
        else:
            self.usingfolder = False
            self.filepath = self.filename
            self.ptcfilepath = self.ptcfilename
        self.particle     = particle
        self.energy       = energy
        self.distribution = distribution
        self.distrkwargs  = {}
        self.nparticles   = nparticles
        self.length       = length
        self.kwargs       = kwargs
        self.figureNr     = 1729 # arbitrary number where figure start from
        
    def CleanMakeRun(self):
        self.Clean()
        self.Make()
        self.Execute()
        self.Compare()
    
    def Clean(self):        
        _os.system("rm -rf "+self.filepath+"*")
        _os.system("rm -rf "+self.foldername+"/output*")
        _os.system("rm -rf "+self.foldername+"/Maxwellian_bend_for_ptc.txt trackone inrays.madx")
        _os.system("rm -rf "+self.foldername+"/test*")
        _os.system("rm -rf "+self.foldername+"/*.log")
        _os.system("rm -rf "+self.foldername+"/*.dat")        
        _os.system("rm -rf "+self.foldername+"/trackone")
        _os.system("rm -rf "+self.foldername+"/inrays.madx")

        # clean and close figures (8 figures in total)
        for i in range(9):
            _plt.close(self.figureNr+i)
        
    def ChangeDistribution(self,distribution='flat',nparticles=10,**kwargs):
        """
        'flat'
        kwargs: mux=0.0, widthx=1e-05, mupx=0.0, widthpx=1e-05, muy=0.0, 
                widthy=0.001, mupy=0.0, widthpy=0.001
        'gauss'
        kwargs: gemx=1e-10, betax=0.1, alfx=0.0, gemy=1e-10, betay=0.1, 
                alfy=0.0, sigmat=1e-12, sigmapt=1e-12
        """
        self.distribution = distribution
        self.distrkwargs  = kwargs
        self.nparticles   = nparticles
        
    def Make(self):
        #type_='drift', foldername=None, particle='e-', energy=1.0,**kwargs) : 
        print('Test> Element type:         ',self.type_)
        print('Test> Destination filepath: ',self.filepath)
        print('Test> kwargs: ', end=' ')
        for k in self.kwargs : 
            print(k+'='+str(self.kwargs[k]), end=' ')

        if self.distribution == 'flat':
            ptc = pymadx.Ptc.FlatGenerator(**self.distrkwargs)
        elif self.distribution == 'gauss':
            ptc = pymadx.Ptc.GaussGenerator(**self.distrkwargs)
        
        ptc.Generate(self.nparticles,self.ptcfilepath)
        
        bb  = pybdsim.Beam.Beam(self.particle,self.energy,'ptc')
        xb  = pymadx.Beam(self.particle,self.energy,'ptc')
        
        bb.SetDistribFileName(self.ptcfilename) 
        xb.SetDistribFileName(self.ptcfilename) 
        
        bm  = pybdsim.Builder.Machine()
        xm  = pymadx.Builder.Machine()
        
        bm.AddBeam(bb)
        xm.AddBeam(xb)
    
        if self.type_ == 'drift' :
            name = 'd1'
            bm.AddDrift(name,length=self.length,**self.kwargs)
            xm.AddDrift(name,length=self.length,**self.kwargs)
        elif self.type_ == 'sbend':
            name = 'sb1'
            bm.AddDipole(name,length=self.length,**self.kwargs)
            xm.AddDipole(name,length=self.length,**self.kwargs)
        elif self.type_ == 'rbend':
            name = 'rb1'
            bm.AddDipole(name,category='rbend',length=self.length,**self.kwargs)
            xm.AddDipole(name,category='rbend',length=self.length,**self.kwargs)
        elif self.type_ == 'hkicker':
            name = 'hk1'
            bm.AddHKicker(name, length=self.length, **self.kwargs)
            xm.AddHKicker(name, length=self.length, **self.kwargs)
        elif self.type_ == 'vkicker':
            name = 'vk1'
            bm.AddVKicker(name, length=self.length, **self.kwargs)
            xm.AddVKicker(name, length=self.length, **self.kwargs)
        elif self.type_ == 'quadrupole' :
            name = 'q1'
            bm.AddQuadrupole(name,length=self.length,**self.kwargs)
            xm.AddQuadrupole(name,length=self.length,**self.kwargs)
        elif self.type_ == 'sextupole' :
            name = 's1'
            bm.AddSextupole(name,length=self.length,**self.kwargs)
            xm.AddSextupole(name,length=self.length,**self.kwargs)
        elif self.type_ == 'solenoid':
            name = 'sl1'
            bm.AddSolenoid(name,length=self.length,**self.kwargs)
            xm.AddSolenoid(name,length=self.length,**self.kwargs)

        bm.AddMarker("theend") # Need a post element marker to sample at, only for bdsim
        
        bm.AddSampler('theend')
        xm.AddSampler('all')

        bm.Write(self.filepath)
        xm.Write(self.filepath)
            
    def Execute(self):
        if self.usingfolder:
            _os.chdir(self.foldername)
        
        _os.system("madx < "+self.filename+".madx > madx.log")
        _os.system("bdsim --file="+self.filename+".gmad --batch --output=combined --outfile='test' > bdsim.log")
        
        if self.usingfolder:
            _os.chdir("../")

    def Compare(self, addPrimaries=True):

        if self.usingfolder:
            _os.chdir(self.foldername)

        bdsimprim = pybdsim.Data.Load("test/test.primaries.txt")
        Bx0 = bdsimprim.X()
        By0 = bdsimprim.Y()
        Bxp0 = bdsimprim.Xp()
        Byp0 = bdsimprim.Yp()
        self.bdsimprimaries = {'x':Bx0,'y':By0,'xp':Bxp0,'yp':Byp0}
        
        bdsim = pybdsim.Data.Load("test/test.txt")
        Bx = bdsim.X()
        By = bdsim.Y()
        Bxp = bdsim.Xp()
        Byp = bdsim.Yp()
        self.bdsimoutput = {'x':Bx,'y':By,'xp':Bxp,'yp':Byp}

        madx = pymadx.Tfs("trackone")
        madx = madx.GetSegment(madx.nsegments) #get the last 'segment' / sampler
        Mx = madx.GetColumn('X')*1e6 # convert from m to um
        My = madx.GetColumn('Y')*1e6 
        Mxp = madx.GetColumn('PX')
        Myp = madx.GetColumn('PY')
        self.ptcoutput = {'x':Mx,'y':My,'xp':Mxp,'yp':Myp}

        fresx  = _np.nan_to_num(Mx - Bx)
        fresy  = _np.nan_to_num(My - By)
        fresx  = _np.nan_to_num(fresx / Mx) #protect against nans for 0 diff
        fresy  = _np.nan_to_num(fresy / My)
        fresxp = _np.nan_to_num(Mxp - Bxp)
        fresyp = _np.nan_to_num(Myp - Byp)
        fresxp = _np.nan_to_num(fresxp / Mxp)
        fresyp = _np.nan_to_num(fresyp / Myp)
        self.residuals = {'x':fresx,'y':fresy,'xp':fresxp,'yp':fresyp}
        
        # 2d plots
        #X vs Y
        _plt.figure(self.figureNr)
        _plt.clf()
        _plt.plot(Mx,My,'b.',label='PTC')
        _plt.plot(Bx,By,'g.',label='BDSIM')
        if addPrimaries:
            _plt.plot(Bx0,By0,'r.',label='BDSIM prim')
        _plt.legend()
        _plt.xlabel(r"x ($\mu$m)")
        _plt.ylabel(r"y ($\mu$m)")
        _plt.title(self.type_)
        _plt.savefig(self.type_+'_xy.pdf')
        _plt.savefig(self.type_+'_xy.png')

        #XP vs YP
        _plt.figure(self.figureNr+1)
        _plt.clf()
        _plt.plot(Mxp,Myp,'b.',label='PTC')
        _plt.plot(Bxp,Byp,'g.',label='BDSIM')
        if addPrimaries:
            _plt.plot(Bxp0,Byp0,'r.',label='BDSIM prim')
        _plt.legend()
        _plt.xlabel(r"x' ($\mu$m)")
        _plt.ylabel(r"y' ($\mu$m)")
        _plt.title(self.type_)
        _plt.savefig(self.type_+'_xpyp.pdf')
        _plt.savefig(self.type_+'_xpyp.png')

        #X vs XP
        _plt.figure(self.figureNr+2)
        _plt.clf()
        _plt.plot(Mx,Mxp,'b.',label='PTC')
        _plt.plot(Bx,Bxp,'g.',label='BDSIM')
        if addPrimaries:
            _plt.plot(Bx0,Bxp0,'r.',label='BDSIM prim')
        _plt.legend()
        _plt.xlabel(r"x ($\mu$m)")
        _plt.ylabel(r"x' (rad)")
        _plt.title(self.type_)
        _plt.savefig(self.type_+'_xxp.pdf')
        _plt.savefig(self.type_+'_xxp.png')

        #Y vs YP
        _plt.figure(self.figureNr+3)
        _plt.clf()
        _plt.plot(My,Myp,'b.',label='PTC')
        _plt.plot(By,Byp,'g.',label='BDSIM')
        if addPrimaries:
            _plt.plot(By0,Byp,'r.',label='BDSIM prim')
        _plt.legend()
        _plt.xlabel(r"y ($\mu$m)")
        _plt.ylabel(r"y' (rad)")
        _plt.title(self.type_)
        _plt.savefig(self.type_+'_yyp.pdf')
        _plt.savefig(self.type_+'_yyp.png')

        # 1d plots
        # x comparison
        f = _plt.figure(self.figureNr+4)
        f.suptitle(self.type_)
        _plt.clf()

        ax1 = f.add_subplot(221)
        ax1.hist(Mx,color='b',label='PTC',histtype='step')
        ax1.hist(Bx,color='g',label='BDSIM',histtype='step')
        if addPrimaries:
            ax1.hist(Bx0,color='r',label='BDSIM prim',histtype='step')
        ax1.legend(fontsize='x-small',loc=0)
        ax1.set_xlabel(r"x ($\mu$m)")
        
        # y comparison
        ax2 = f.add_subplot(222)
        ax2.hist(My,color='b',label='PTC',histtype='step')
        ax2.hist(By,color='g',label='BDSIM',histtype='step')
        if addPrimaries:
            ax2.hist(By0,color='r',label='BDSIM prim',histtype='step')
        ax2.legend(fontsize='x-small',loc=0)
        ax2.set_xlabel(r"y ($\mu$m)")

        # xp comparison
        ax3 = f.add_subplot(223)
        ax3.hist(Mxp,color='b',label='PTC',histtype='step')
        ax3.hist(Bxp,color='g',label='BDSIM',histtype='step')
        if addPrimaries:
            ax3.hist(Bxp0,color='r',label='BDSIM prim',histtype='step')
        ax3.legend(fontsize='x-small',loc=0)
        ax3.set_xlabel(r"x' (rad)")

        # yp comparison
        ax4 = f.add_subplot(224)
        ax4.hist(Myp,color='b',label='PTC',histtype='step')
        ax4.hist(Byp,color='g',label='BDSIM',histtype='step')
        if addPrimaries:
            ax4.hist(Byp0,color='r',label='BDSIM prim',histtype='step')
        ax4.legend(fontsize='x-small',loc=0)
        ax4.set_xlabel(r"y' (rad)")
        
        _plt.savefig(self.type_+'_hist.pdf')
        _plt.savefig(self.type_+'_hist.png')

        # residuals in one plot
        f = _plt.figure(self.figureNr+8)
        _plt.clf()
        
        axX = f.add_subplot(221)
        axX.hist(Mx,weights=fresx,bins=100,ec='b')
        axX.set_xlabel(r'X ($\mu$m)')
        axX.set_ylabel('Fractional Residuals')
        
        axY = f.add_subplot(222)
        axY.hist(My,weights=fresy,bins=100,ec='b')
        axY.set_xlabel(r'Y ($\mu$m)')
        axY.set_ylabel('Fractional Residuals')
        
        axXp = f.add_subplot(223)
        axXp.hist(Mxp*1e3,weights=fresxp,bins=100,ec='b')
        axXp.set_xlabel('Xp (mrad)')
        axXp.set_ylabel('Fractional Residuals')

        axYp = f.add_subplot(224)
        axYp.hist(Myp*1e3,weights=fresyp,bins=100,ec='b')
        axYp.set_xlabel('Yp (mrad)')
        axYp.set_ylabel('Fractional Residuals')

        _plt.subplots_adjust(left=0.15,right=0.95,top=0.95,wspace=0.39,hspace=0.25)
        _plt.savefig(self.type_+'_residuals.pdf')
        _plt.savefig(self.type_+'_residuals.png')
        
        #emittance
        r = robdsim.robdsimOutput("test.root")
        r.CalculateOpticalFunctions("optics.dat")
        d = pybdsim.Data.Load("optics.dat")
        print('Horizontal emittance bdsim (before,after) ',d.Emitt_x())
        print('Vertical emittance bdsim (before,after) ',d.Emitt_y())

        if self.usingfolder:
            _os.chdir("../")



import pickle as _pkl
import pylab as _pl
import pymad8 as _m8
import pybdsim as _pybdsim
import matplotlib.pyplot as _plt
import numpy as _np
from os.path import isfile as _isfile
from matplotlib.backends.backend_pdf import PdfPages as _PdfPages
import datetime as _datetime

# Predefined dicts for making the standard plots,
# format = (mad8_optical_var_name, bdsim_optical_var_name, bdsim_optical_var_error_name, legend_name)

_BETA = {"bdsimdata": ("Beta_x", "Beta_y"),
         "bdsimerror": ("Sigma_Beta_x", "Sigma_Beta_y"),
         "mad8": ("BETX", "BETY"),
         "legend": (r'$\beta_{x}$', r'$\beta_{y}$'),
         "xlabel": "S / m",
         "ylabel": r"$\beta_{x,y}$ / m",
         "title": "Beta"}

_ALPHA = {"bdsimdata": ("Alpha_x", "Alpha_y"),
          "bdsimerror": ("Sigma_Alpha_x", "Sigma_Alpha_y"),
          "mad8": ("ALPHX", "ALPHY"),
          "legend": (r'$\alpha_{x}$', r'$\alpha_{y}$'),
          "xlabel": "S / m",
          "ylabel": r"$\alpha_{x,y}$ / m",
          "title": "Alpha"}

_DISP = {"bdsimdata": ("Disp_x", "Disp_y"),
         "bdsimerror": ("Sigma_Disp_x", "Sigma_Disp_y"),
         "mad8": ("DX", "DY"),
         "legend": (r"$\eta_{x}$", r"$\eta_{y}$"),
         "xlabel": "S / m",
         "ylabel": r"$\eta_{x,y} / m$",
         "title": "Dispersion"}

_DISP_P = {"bdsimdata": ("Disp_xp", "Disp_yp"),
           "bdsimerror": ("Sigma_Disp_xp", "Sigma_Disp_yp"),
           "mad8": ("DPX", "DPY"),
           "legend": (r"$\eta_{p_x}$", r"$\eta_{p_x}$"),
           "xlabel": "S / m",
           "ylabel": r"$\eta_{p_{x},p_{y}}$ / m",
           "title": "Momentum_Dispersion"}

_SIGMA = {"bdsimdata": ("Sigma_x", "Sigma_y"),
          "bdsimerror": ("Sigma_Sigma_x", "Sigma_Sigma_y"),
          "mad8": ("", ""),
          "legend": (r"$\sigma_{x}$", r"$\sigma_{y}$"),
          "xlabel": "S / m",
          "ylabel": r"$\sigma_{x,y}$ / m",
          "title": "Sigma"}

_SIGMA_P = {"bdsimdata": ("Sigma_xp", "Sigma_yp"),
            "bdsimerror": ("Sigma_Sigma_xp", "Sigma_Sigma_yp"),
            "mad8": ("", ""),
            "legend": (r"$\sigma_{xp}$", r"$\sigma_{yp}$"),
            "xlabel": "S / m",
            "ylabel": r"$\sigma_{xp,yp}$ / rad",
            "title": "SigmaP"}

_MEAN = {"bdsimdata": ("Mean_x", "Mean_y"),
         "bdsimerror": ("Sigma_Mean_x", "Sigma_Mean_y"),
         "mad8": ("X", "Y"),
         "legend": (r"$\overline{x}$", r"$\overline{y}$"),
         "xlabel": "S / m",
         "ylabel": r"$\bar{x,y}$ / m",
         "title": "Mean"}

_EMITT = {"bdsimdata": ("Emitt_x", "Emitt_y"),
          "bdsimerror": ("Sigma_Emitt_x", "Sigma_Emitt_y"),
          "mad8": ("", ""),
          "legend": (r"$\epsilon_x$", r"$\epsilon_y$"),
          "xlabel": "S / m",
          "ylabel": r"$\epsilon_{x,y}$ / m",
          "title": "Emittance"}


# use closure to avoid tonnes of boilerplate code as happened with MadxBdsimComparison.py
def _make_plotter(plot_info_dict):
    def f_out(mad8opt, bdsopt, beamParams, functions=None, postfunctions=None, survey=None, figsize=(9, 5), xlim=(0, 0), **kwargs):

        # Get the initial N for the bdsim
        N = str(int(bdsopt['Npart'][0]))  # number of primaries.

        # labels for plot legends
        mad8legendx = r'MAD8 ' + plot_info_dict['legend'][0]
        mad8legendy = r'MAD8 ' + plot_info_dict['legend'][1]
        bdslegendx = r'BDSIM ' + plot_info_dict['legend'][0] + ' ; N = ' + N
        bdslegendy = r'BDSIM ' + plot_info_dict['legend'][1] + ' ; N = ' + N

        # mad8 data from correct source
        if plot_info_dict["title"] == "Sigma":
            mad8opt.calcBeamSize(beamParams['ex'], beamParams['ey'], beamParams['esprd'])
            mad8Xdata = mad8opt.getColumnsByKeys('SIGX')
            mad8Ydata = mad8opt.getColumnsByKeys('SIGY')

            mad8s = mad8opt.getColumnsByKeys('S')
            mad8legendx += '(calculated)'
            mad8legendy += '(calculated)'
        elif plot_info_dict["title"] == "SigmaP":
            mad8opt.calcBeamSize(beamParams['ex'], beamParams['ey'], beamParams['esprd'])
            mad8Xdata = mad8opt.getColumnsByKeys('SIGXP')
            mad8Ydata = mad8opt.getColumnsByKeys('SIGYP')

            mad8s = mad8opt.getColumnsByKeys('S')
            mad8legendx += '(calculated)'
            mad8legendy += '(calculated)'
        elif plot_info_dict["title"] == "Emittance":
            emitX, emitY = _CalculateEmittance(mad8opt, beamParams)
            mad8Xdata = emitX
            mad8Ydata = emitY
            mad8s = mad8opt.getColumnsByKeys('S')
        else:
            mad8Xdata = mad8opt.getColumnsByKeys(plot_info_dict['mad8'][0])
            mad8Ydata = mad8opt.getColumnsByKeys(plot_info_dict['mad8'][1])
            mad8s = mad8opt.getColumnsByKeys('S')

        # the figure
        plot = _plt.figure(plot_info_dict["title"], figsize=figsize, **kwargs)

        # mad8 plot
        _plt.plot(mad8s, mad8Xdata, 'b--', label=mad8legendx)
        _plt.plot(mad8s, mad8Ydata, 'g--', label=mad8legendy)

        # bds plot
        _plt.errorbar(bdsopt['S'], bdsopt[plot_info_dict['bdsimdata'][0]], bdsopt[plot_info_dict['bdsimerror'][0]],
                      label=bdslegendx, capsize=3, ls='', marker='x', color='b', **kwargs)
        _plt.errorbar(bdsopt['S'], bdsopt[plot_info_dict['bdsimdata'][1]], bdsopt[plot_info_dict['bdsimerror'][1]],
                      label=bdslegendy, capsize=3, ls='', marker='x', color='g', **kwargs)

        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(plot_info_dict['ylabel'])
        axes.set_xlabel(plot_info_dict['xlabel'])
        axes.legend(loc='best')
        axes.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))

        if survey is None:
            survey = mad8opt
        _CallUserFigureFunctions(functions)
        _AddSurvey(plot, survey)
        _CallUserFigureFunctions(postfunctions)

        plot.sca(plot.axes[0])
        _plt.show(block=False)

        if xlim != (0, 0):
            _plt.xlim(xlim)
        _plt.show(block=False)
        return plot

    return f_out


PlotBeta = _make_plotter(_BETA)
PlotAlpha = _make_plotter(_ALPHA)
PlotDisp = _make_plotter(_DISP)
PlotDispP = _make_plotter(_DISP_P)
PlotSigma = _make_plotter(_SIGMA)
PlotSigmaP = _make_plotter(_SIGMA_P)
PlotMean = _make_plotter(_MEAN)
PlotEmitt = _make_plotter(_EMITT)


def _CalculateEmittance(mad8opt, beamParams):
    emitX0 = beamParams['ex']
    emitY0 = beamParams['ey']
    particle = beamParams['particle']
    if particle == 'electron' or particle == 'positron':
        mass = 0.5109989461
    elif particle == 'proton':
        mass = 938.2720813
    else:  # default is mad8 default particle mass.
        mass = 0.5109989461

    e = mad8opt.getColumnsByKeys('E')
    rgamma = e / (mass / 1e3)
    rbeta = _np.sqrt(1 - 1.0 / rgamma ** 2)

    emitXN0 = emitX0 * rgamma[0] * rbeta[0]
    emitYN0 = emitY0 * rgamma[0] * rbeta[0]

    emitX = emitXN0 / (rbeta * rgamma)
    emitY = emitYN0 / (rbeta * rgamma)
    return emitX, emitY


def _CallUserFigureFunctions(functions):
    if isinstance(functions, list):
        for function in functions:
            if callable(function):
                function()
    elif callable(functions):
        functions()


def _AddSurvey(figure, survey):
    if survey is None:
        return
    else:
        _m8.Plot.AddMachineLatticeToFigure(figure, survey)


def Mad8VsBDSIM(twiss, bdsim, survey=None, functions=None, postfunctions=None, figsize=(10, 5), xlim=(0, 0),
                saveAll=True, outputFileName=None, particle="electron", energySpread=1e-4, ex=1e-8, ey=1e-8):
    """ Compares Mad8 and BDSIM optics variables.

    +-----------------+---------------------------------------------------------+
    | **Parameters**  | **Description**
    +-----------------+---------------------------------------------------------+
    | twiss           | Mad8 twiss file                                         |
    +-----------------+---------------------------------------------------------+
    | bdsim           | Optics root file (from rebdsimOptics or rebdsim).       |
    +-----------------+---------------------------------------------------------+
    | functions       | Hook for users to add their functions that are called   |
    |                 | immediately prior to the addition of the plot. Use a    |
    |                 | lambda function to add functions with arguments. Can    |
    |                 | be a function or a list of functions.                   |
    +-----------------+---------------------------------------------------------+
    | figsize         | Figure size for all figures - default is (12,5)         |
    +-----------------+---------------------------------------------------------+
    | xlim            | Set xlimit for all figures                              |
    +-----------------+---------------------------------------------------------+
    | particle        | Beam particle type to determine particle mass, required |
    |                 | for beam size calculation - default is electron.        |
    +-----------------+---------------------------------------------------------+
    | energySpread    | Energy spread used in beam size calculation - default   |
    |                 | is 1e-4.                                                |
    +-----------------+---------------------------------------------------------+
    | ex / ey         | Horizontal / vertical emittance used in beam size       |
    |                 | calculation - default is 1e-8.                          |
    +-----------------+---------------------------------------------------------+
    """

    if not _isfile(twiss):
        raise IOError("File not found: ", twiss)
    if isinstance(bdsim, str) and not _isfile(bdsim):
        raise IOError("File not found: ", bdsim)

    fname = _pybdsim.Data.GetFileName(bdsim)  # cache file name
    if fname == "":
        fname = "optics_report"

    # load mad8 optics and bdsim optics
    mad8opt = _m8.Output(twiss)
    bdsinst = _pybdsim.Data.CheckItsBDSAsciiData(bdsim)
    bdsopt = _GetBDSIMOptics(bdsinst)

    # parameters required for calculating beam sizes, not written in mad8 output so have to supply manually.
    beamParams = {'esprd': energySpread, 'particle': particle, 'ex': ex, 'ey': ey}

    # make plots
    # energy and npart plotted with individual methods
    figures = [PlotBeta(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotAlpha(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotDisp(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotDispP(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotSigma(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotSigmaP(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotEnergy(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotMean(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotEmitt(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey),
               PlotNParticles(mad8opt, bdsopt, beamParams, functions=functions, postfunctions=postfunctions, figsize=figsize, xlim=xlim, survey=survey)]

    if saveAll:
        tfsname = repr(twiss)
        bdsname = repr(bdsinst)
        output_filename = "optics-report.pdf"
        if outputFileName is not None:
            output_filename = outputFileName
            if not output_filename.endswith('.pdf'):
                output_filename += ".pdf"
        else:
            output_filename = fname.replace('.root', '')
            output_filename += ".pdf"
        # Should have a more descriptive name really.
        with _PdfPages(output_filename) as pdf:
            for figure in figures:
                pdf.savefig(figure)
            d = pdf.infodict()
            d['Title'] = "{} (MAD8) VS {} (BDSIM) Optical Comparison".format(tfsname, bdsname)
            d['CreationDate'] = _datetime.datetime.today()
        print("Written ", output_filename)
    return mad8opt


def PlotEnergy(mad8opt, bdsopt, beamParams, survey=None, functions=None, postfunctions=None, figsize=(12, 5), xlim=(0, 0)):
    N = str(int(bdsopt['Npart'][0]))  # number of primaries.
    energyPlot = _plt.figure('Energy', figsize)

    # one missing energy due to initial
    _plt.plot(mad8opt.getColumnsByKeys('S'), mad8opt.getColumnsByKeys('E'), 'b--', label=r'MAD8 $E$')

    _plt.errorbar(bdsopt['S'], bdsopt['Mean_E'], yerr=bdsopt['Sigma_Mean_E'], label=r'BDSIM $E$' + ' ; N = ' + N, marker='x', ls='', color='b')

    axes = _plt.gcf().gca()
    axes.set_ylabel('Energy / GeV')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    if survey is None:
        survey = mad8opt
    _CallUserFigureFunctions(functions)
    _AddSurvey(energyPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    if xlim != (0, 0):
        _plt.xlim(xlim)

    energyPlot.sca(energyPlot.axes[0])

    _plt.show(block=False)
    return energyPlot


def PlotNParticles(mad8opt, bdsopt, beamParams, survey=None, functions=None, postfunctions=None, figsize=(12, 5), xlim=(0, 0)):
    npartPlot = _plt.figure('NParticles', figsize)

    _plt.plot(bdsopt['S'], bdsopt['Npart'], 'k-', label='BDSIM N Particles')
    _plt.plot(bdsopt['S'], bdsopt['Npart'], 'k.')
    axes = _plt.gcf().gca()
    axes.set_ylabel(r'N Particles')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    if survey is None:
        survey = mad8opt
    _CallUserFigureFunctions(functions)
    _AddSurvey(npartPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    if xlim != (0, 0):
        _plt.xlim(xlim)

    npartPlot.sca(npartPlot.axes[0])

    _plt.show(block=False)
    return npartPlot


def _GetBDSIMOptics(optics):
    """Takes a BDSAscii instance. Return a dictionary of lists matching the variable with the list of values."""

    optvars = {}
    for variable in optics.names:
        datum = getattr(optics, variable)()
        optvars[variable] = datum
    return optvars



from .. import Builder as _Builder
import pybdsim._General as _General
import pybdsim.Data as _Data

_ElementModifier = _Builder.ElementModifier

def ElegantParam2GmadStrength(paramfile, outputfilename,
                              existingmachine = None,
                              verbose         = False,
                              flipmagnets     = False,
                              linear          = False,
                              ignoreZeroLengthItems = True):
    """
    Use an Elegant 'parameters' file containing details about each component to generate a 
    strength (only) BDSIM GMAD file to be used with an existing lattice.
    
    +------------------+----------------------------------------------------------------+
    | existingmachine  | either a list or dictionary with names of elements to prepare. |
    +------------------+----------------------------------------------------------------+
    | flipmagnet       | whether to flip k values for  negatively charged particles.    |
    +------------------+----------------------------------------------------------------+
    | linear           | only use linear strengths, k2 and higher set to 0.             |
    +------------------+----------------------------------------------------------------+

    """
    params = _Data.LoadSDDSColumnsToDict(paramfile)
    paramsd = _Data.SDDSBuildParameterDicts(params)

    if existingmachine == None:
        existingnames = []
    elif type(existingmachine) == list:
        existingnames = existingmachine
    elif type(existingmachine) == dict:
        existingnames = list(existingmachine.keys()) #dictionary
    else:
        raise ValueError("Unsuitable type of existing machine")

    newStrengths = []

    for name,paramDict in paramsd.items():
        if _WillIgnoreItem(paramDict, ignoreZeroLengthItems):
            continue
        
        # remove special characters like $, % etc 'reduced' name - rname:
        rname = _General.PrepareReducedName2(name)

        # generate elementmodifier with approprate name to match one
        # already used in existing machine
        nameToUse = name if name in existingnames else rname
        a = _GenerateElementModifier(paramDict, nameToUse, verbose, flipmagnets, linear)
        
        if verbose:
            print(a)
        if a != None:
            newStrengths.append(a)

    #write output
    if not outputfilename.endswith('.gmad'):
        outputfilename += '.gmad'
    f = open(outputfilename, 'w')
    for strength in newStrengths:
        f.write(str(strength))
    f.close()


def _GenerateElementModifier(item, nameToUse, verbose=False, flipmagnets=False, linear=False):
    """
    Generate an Builder.ElementModifier instance based on the particular
    element / magnet type.  

    Takes second argument of nameToUse to match whatever the name as been
    changed to.

    This function doesn't do any key checking for the dictionary as that should
    be done by MadxTfs2GmadStrength
    """
    category = item['KEYWORD']
    factor   = -1 if flipmagnets else 1  #flipping magnets
    
    a = None
    if category == 'KQUAD':
        newk1 = item['K1'] * factor
        a = _ElementModifier(nameToUse, k1=newk1)
    else:
        pass # just keep a = None and return that

    #TBC for other types

    if verbose:
        print(('New Strength: ', a))
        if a is None:
            print(('Unsupported type: ', category))
    
    return a


def _WillIgnoreItem(paramDict, ignoreZeroLengthItems=False, ignoreableThinElements=[]):
    kw = paramDict['KEYWORD']

    if kw in ['CHARGE', 'ENERGY', 'FLOOR', 'MALIGN', 'MARK', 'MONI', 'WATCH']:
        return True
    else:
        if 'L' in paramDict:
            l = paramDict['L']
            zeroLength = l < 1e-9
            if zeroLength and ignoreZeroLengthItems:
                return True
            if zeroLength and kw in ignoreableThinElements:
                return True
    return False
        



"""
Work in progress.

"""

from . import Data as _Data
import pylab as _pl

class Helper:
    """
    To help locate objects in the BDSIM visualisation, requires a BDSIM survey file
    """
    def __init__(self, surveyFileName):
        self.survey      = _Data.Load(surveyFileName)
        self.x           = self.survey.X()
        self.y           = self.survey.Y()
        self.z           = self.survey.Z()
        self.worldCentre = self.getWorldCentre()
        

    def getWorldCentre(self, type="linear"):
        """
        Returns the center in world coordinates of the centre of the visualisation space
        """
        if type == "linear":
            xCentre = (self.x[0]+self.x[-1])/2.0
            yCentre = (self.y[0]+self.y[-1])/2.0
            zCentre = (self.z[0]+self.z[-1])/2.0
        elif type == "circular":
            xCentre = (self.x.max()+self.x.min())/2.0
            yCentre = (self.y.min()+self.y.max())/2.0
            zCentre = (self.z.max()+self.z.min())/2.0

       
        print("Visualisation.Helper.getWorldCentre>", xCentre, yCentre, zCentre)
        return _pl.array([xCentre, yCentre, zCentre])

    def findComponentCoords(self, componentName):
        """
        Returns the XYZ coordinates of a component relative to the centre
        """
        names = self.survey.Name()

        matchingNames = []
        matchingIdx   = []

        idx = 0
        # loop over names 
        for name in names : 
            if name.rfind(componentName) != -1 : 
                print('Visualisation.Helper.findComponentCoords>',name,idx)
                matchingNames.append(name)
                matchingIdx.append(idx)
            idx = idx+1
            
        i = 0
        for match in matchingNames : 
            idx = matchingIdx[matchingNames.index(match)]
            x   = self.x[idx]
            y   = self.y[idx]
            z   = self.z[idx]
            v   = _pl.array([x,y,z]) 
            vp  = v - self.worldCentre
            print(match,idx,v,vp)
            
    def draw(self) : 
        """
        Quick survey drawing for diagnostic reasons.
        """
        _pl.subplot(2,1,1)
        _pl.plot(self.z, self.x)
        _pl.plot(self.worldCentre[2], self.worldCentre[0],"+")
        _pl.xlabel("Z [m]")
        _pl.ylabel("X [m]")
        _pl.subplot(2,1,2)
        _pl.plot(self.z, self.y)
        _pl.plot(self.worldCentre[2], self.worldCentre[1],"+")
        _pl.xlabel("Z [m]")
        _pl.ylabel("Y [m]")




def MadxVsMadx(first, second, first_name=None,
               second_name=None, **kwargs):
    """
    Display all the optical function plots for the two input optics files.
    """
    import pymadx as _pymadx
    _pymadx.Compare.MadxVsMadx(first, second, first_name, second_name, **kwargs)



# pybdsim.Analysis - analysis scripts for bdsim output
# Version 1.0
# L. Nevay, S.T.Boogert
# laurie.nevay@rhul.ac.uk

#this module does not feature directly in the pybdsim import
#but instead acts as a holder for the analysis class to
#keep the __init__ file very clean.

import Data
import numpy as _np
import Plot
import Constants

class Analysis:
    """
    Analysis class for bdsim output

    one instance of the class per output file

    Analysis('../../path/to/my/output.txt')
    
    It will also look for '../../path/to/my/output.eloss.txt' 
    beside the output file

    """
    def __init__(self,filepath):
        self.data,self.dataarray,self.keyslist = Data.LoadOld(filepath)
        self.filepath = filepath
        self.filename = filepath.split('/')[-1]
        self.plots     = []
        self._plottitle = self.filename.split('.')[0]
        self.ndata     = len(self.data['X'])
        
    def GroupBy(self,variable='Z'):
        """
        GroupBy(variable='Z')
        
        create instance.datagrouped dictionary
        finds unique values of variable and groups all the 
        data that has that value of variable

        e.g.
        GroupBy()   # default is 'Z'
        instance.datagrouped is dict with:
        0.000: array of (nparticles x allother variables)
        1.202: similar but different number of particles maybe
        ... etc
        """
        self.datagrouped = Data.AsciiData()
        
        #find unique values of variable
        uniquevalues = sorted(list(set(_np.round(self.data[variable],2))))

        #remove the variable from the subset
        #find it's index in keys list
        indexofvariabletoremove = self.keyslist.index(variable)
        for value in uniquevalues:
            mask      = _np.round(self.data[variable],2) == value
            dcopy     = self.dataarray[mask]
            dcopydict = dict(zip(self.keyslist,[dcopy[:,i] for i in range(_np.shape(dcopy)[1])]))

            #dcopydict = Data.AsciiData(zip(self.data.keyslist,[dcopy[:,i] for i in range(_np.shape(dcopy)[1])]))
            dcopydict['nparticles'] = _np.shape(dcopy)[0]
            self.datagrouped[value] = dcopydict
        self.keysgrouped = list(_np.sort(list(self.datagrouped.keys())))

    def GenerateSigmas(self):
        if hasattr(self,'datagrouped') == False:
            self.GroupBy()
        
        z = self.keysgrouped
        sx,sy = [],[]

        for i in range(len(self.keysgrouped)):
            sx.append(_np.std(self.datagrouped[self.keysgrouped[i]]['X']))
            sy.append(_np.std(self.datagrouped[self.keysgrouped[i]]['Y']))
        self.simpledata = {'sx':sx,'sy':sy,'z':z}

    def SortBy(self,variable='Z'):
        pass

    def PlotXXp(self):
        p1 = Plot.PlotXY(self.data['X'],self.data['Xp'],'X ('+self.data.units['X']+')','Xp ('+self.data.units['Xp']+')',self._plottitle)
        self.plots.append(p1)

    def PlotYYp(self):
        p1 = Plot.PlotXY(self.data['Y'],self.data['Yp'],'Y ('+self.data.units['Y']+')','Yp ('+self.data.units['Y']+')',self._plottitle)
        self.plots.append(p1)

    def PlotSingleVariable(self,variable,xlabel='',ylabel=''):
        p1 = Plot.PlotY(self.data[variable],xlabel,ylabel,self._plottitle)
        self.plots.append(p1)

    def ControlPlotAll(self):
        p1 = Plot.ControlPlot(self.data,self._plottitle)
        self.plots.append(p1)
        
    def PlotEnergyHist(self,nbins=20,**kwargs):
        p1 = Plot.Histogram(self.data['E'],'Energy ('+self.data.units['E']+')',self._plottitle,nbins,**kwargs)
        self.plots.append(p1)

    




import pybdsim

d = pybdsim.Data.Load("r2-ana.root")

h1 = d.histograms1dpy['Event/PerEntryHistograms/Q2Neutrons']
h2 = d.histograms1dpy['Event/PerEntryHistograms/Q2Gammas']

binWidth = str(round(h1.xwidths[0],1))

h1c, h2c, r = pybdsim.Plot.Histogram1DRatio(h1,h2, yLogScale=True,
                              xlabel="Total Energy (GeV)",
                              ylabel="Fraction / Event / "+binWidth+" GeV",
                              title="Comparison of Fluxes",
                              histogram1Colour='r',
                              histogram2Colour='b',
                              ratioColour='k')






"""
General utilities for day to day housekeeping
"""

import glob as _glob
import os as _os
import re as _re
import numpy as _np

def GenUniqueFilename(filename):
    i = 1
    parts = filename.split('.')
    basefilename = parts[0]
    if len(parts) > 1:
        extension = '.' + parts[1]
    else:
        extension = ''
    while _os.path.exists(filename) :
        filename = basefilename+'_'+str(i)+extension
        i = i + 1
    return filename

def Chunks(l, n):
    """
    Yield successive n-sized chunks from l.
    """
    return [l[i:i+n] for i in range(0,len(l),n)]

def NearestEvenInteger(number):
    number = int(number)
    return number + number%2

def Cast(string):
    """
    Cast(string)
    
    tries to cast to a (python)float and if it doesn't work, 
    returns a string

    """
    try:
        return float(string)
    except ValueError:
        return string

def IsFloat(stringtotest):
    try:
        float(stringtotest)
        return True
    except ValueError:
        return False

def PrepareReducedName(elementname):
    """
    Only allow alphanumeric characters and '_'
    """
    rname = _re.sub('[^a-zA-Z0-9_]+','',elementname)
    return rname

def PrepareReducedName2(elementname):
    """
    Only allow alphanumeric characters and '_'
    """
    rname = _re.sub('[^a-zA-Z0-9_\.]+','',elementname)
    return rname

def GetLatestFileFromDir(dirpath='', extension='*'):
    return max(_glob.iglob(dirpath+extension), key=_os.path.getctime)

def IsROOTFile(path):
    """Check if input is a ROOT file."""
    return path[-4:] == "root"



"""
Optical calculations. Work in progress.

"""

import itertools as _itertools
from logging import warning as _warning
import warnings as _warnings
import pandas as _pd
import numpy as _np
from typing import Optional as _Optional
from typing import List as _List
from typing import Dict as _Dict
from typing import Tuple as _Tuple

from pint import UnitRegistry as _UnitRegistry

_ureg = _UnitRegistry(on_redefinition="ignore")
_ureg.define('electronvolt = e * volt = eV')


def _get_matrix_elements_block(m: _pd.DataFrame, twiss: _Dict, block: int = 1) -> _Tuple:
    """Extract parameters from the DataFrame."""
    p = 1 if block == 1 else 3
    v = 1 if block == 1 else 2
    r11: _pd.Series = m[f"R{p}{p}"]
    r12: _pd.Series = m[f"R{p}{p + 1}"]
    r21: _pd.Series = m[f"R{p + 1}{p}"]
    r22: _pd.Series = m[f"R{p + 1}{p + 1}"]
    alpha: float = twiss[f"ALPHA{v}{v}"]
    beta: float = twiss[f"BETA{v}{v}"]
    gamma: float = twiss[f"GAMMA{v}{v}"]

    return r11, r12, r21, r22, alpha, beta, gamma

class TwissException(Exception):
    """Exception raised for errors in the Twiss module."""

    def __init__(self, m):
        self.message = m

class Twiss:
    def __init__(self,
                 samplers_data: _Dict[str, _pd.DataFrame] = None,
                 first_sampler: str = None,
                 last_sampler: str = None
                 ):
        """
        Args:
            samplers_data: Dictionary of Dataframe that contains all the samplers. {sampler's name: dataFrame}.
                           The following coordinates must be in the DataFrame and in this order
                           ['x', 'xp', 'y', 'yp', 'S', 'p']
            first_sampler: First element where to compute the optics
            last_sampler: Last element where to compute the optics

        Returns:

        """
        samplers_name = list(samplers_data)
        if first_sampler is None:
            first_sampler = samplers_name[0]
        if last_sampler is None:
            last_sampler = samplers_name[-1]
        if first_sampler not in samplers_name:
            raise TwissException(f"{first_sampler} not in samplers_data")
        if last_sampler not in samplers_name:
            raise TwissException(f"{last_sampler} not in samplers_data")
        self.samplers_data = dict(_itertools.islice(samplers_data.items(),
                                                    samplers_name.index(first_sampler),
                                                    samplers_name.index(last_sampler)+1))

    def __call__(self,
                 twiss_init: _Optional[_Dict] = None,
                 with_phase_unrolling: bool = True,
                 offsets: _List = None, # -> Must be computed on the fly
                 ) -> _pd.DataFrame:
        """
        Uses a step-by-step transfer matrix to compute the Twiss parameters (uncoupled). The phase advance and the
        determinants of the jacobians are computed as well.

        Args:
            matrix: the input step-by-step transfer matrix

        Returns:
            the same DataFrame as the input, but with added columns for the computed quantities.
        """
        matrix = self.compute_matrix_for_twiss(offsets)

        if twiss_init is None:
            twiss_init = self.compute_periodic_twiss(matrix)

        matrix['BETA11'] = self.compute_beta_from_matrix(matrix, twiss_init)
        matrix['BETA22'] = self.compute_beta_from_matrix(matrix, twiss_init, plane=2)
        matrix['ALPHA11'] = self.compute_alpha_from_matrix(matrix, twiss_init)
        matrix['ALPHA22'] = self.compute_alpha_from_matrix(matrix, twiss_init, plane=2)
        matrix['GAMMA11'] = self.compute_gamma_from_matrix(matrix, twiss_init)
        matrix['GAMMA22'] = self.compute_gamma_from_matrix(matrix, twiss_init, plane=2)
        matrix['MU1'] = self.compute_mu_from_matrix(matrix, twiss_init)
        matrix['MU2'] = self.compute_mu_from_matrix(matrix, twiss_init, plane=2)
        matrix['DET1'] = self.compute_jacobian_from_matrix(matrix, twiss_init)
        matrix['DET2'] = self.compute_jacobian_from_matrix(matrix, twiss_init, plane=2)
        matrix['DISP1'] = self.compute_dispersion_from_matrix(matrix, twiss_init)
        matrix['DISP2'] = self.compute_dispersion_prime_from_matrix(matrix, twiss_init)
        matrix['DISP3'] = self.compute_dispersion_from_matrix(matrix, twiss_init, plane=2)
        matrix['DISP4'] = self.compute_dispersion_prime_from_matrix(matrix, twiss_init, plane=2)

        def phase_unrolling(phi):
            """TODO"""
            if phi[0] < 0:
                phi[0] += 2 * _np.pi
            for i in range(1, phi.shape[0]):
                if phi[i] < 0:
                    phi[i] += 2 * _np.pi
                if phi[i - 1] - phi[i] > 0.5:
                    phi[i:] += 2 * _np.pi
            return phi

        try:
            from numba import njit
            phase_unrolling = njit(phase_unrolling)
        except ModuleNotFoundError:
            pass

        if with_phase_unrolling:
            matrix['MU1U'] = phase_unrolling(matrix['MU1'].values)
            matrix['MU2U'] = phase_unrolling(matrix['MU2'].values)

        return matrix


    def compute_matrix_for_twiss(self, offsets) -> _pd.DataFrame:

        normalization = 2 * offsets
        step_by_step_matrix = _pd.DataFrame()
        for sampler_name, data_sampler in self.samplers_data.items():
            if data_sampler.empty or len(data_sampler) != 11:
                raise TwissException(f"Sampler {sampler_name} is empty or particles are missing. "
                                     f"Size of sampler {sampler_name} = {len(data_sampler)}")

            p0 = data_sampler['p'][0]
            data_sampler['dpp'] = (data_sampler['p'] - p0) / p0

            output_coordinates = data_sampler.drop(columns=['p', 'S'])
            m = output_coordinates.values
            matrix = {
                f'R{j + 1}{i + 1}': ((m[i + 1, j] - m[i + 1 + 5, j]) / normalization[i]) + m[0, j]
                for j in range(0, 5)
                for i in range(0, 5)
            }

            matrix['S'] = data_sampler.iloc[0]['S']
            matrix['X0'] = data_sampler.iloc[0]['x']
            matrix['XP0'] = data_sampler.iloc[0]['xp']
            matrix['Y0'] = data_sampler.iloc[0]['y']
            matrix['YP0'] = data_sampler.iloc[0]['yp']
            step_by_step_matrix = step_by_step_matrix.append(_pd.DataFrame.from_dict(matrix,
                                                                                     orient='index',
                                                                                     columns=[sampler_name]).T)
        return step_by_step_matrix

    @staticmethod
    def compute_alpha_from_matrix(m: _pd.DataFrame, twiss: _Dict, plane: int = 1) -> _pd.Series:
        """
        Computes the Twiss alpha values at every step of the input step-by-step transfer matrix.

        Args:
            m: the step-by-step transfer matrix for which the alpha values should be computed
            twiss: the initial Twiss values
            plane: an integer representing the block (1 or 2)

        Returns:
            a Pandas Series with the alpha values computed at all steps of the input step-by-step transfer matrix
        """
        r11, r12, r21, r22, alpha, beta, gamma = _get_matrix_elements_block(m, twiss, plane)
        return -r11 * r21 * beta + (r11 * r22 + r12 * r21) * alpha - r12 * r22 * gamma

    @staticmethod
    def compute_beta_from_matrix(m: _pd.DataFrame, twiss: _Dict, plane: int = 1,
                                 strict: bool = False) -> _pd.Series:
        """
        Computes the Twiss beta values at every step of the input step-by-step transfer matrix.

        Args:
            m: the step-by-step transfer matrix for which the beta values should be computed
            twiss: the initial Twiss values
            plane: an integer representing the block (1 or 2)
            strict: flag to activate the strict mode: checks and ensures that all computed beta are positive

        Returns:
            a Pandas Series with the beta values computed at all steps of the input step-by-step transfer matrix
        """
        r11, r12, r21, r22, alpha, beta, gamma = _get_matrix_elements_block(m, twiss, plane)
        _ = r11 ** 2 * beta - 2.0 * r11 * r12 * alpha + r12 ** 2 * gamma
        if strict:
            assert (_ > 0).all(), "Not all computed beta are positive."
        return _

    @staticmethod
    def compute_gamma_from_matrix(m: _pd.DataFrame, twiss: _Dict, plane: int = 1) -> _pd.Series:
        """
        Computes the Twiss gamma values at every step of the input step-by-step transfer matrix.

        Args:
            m: the step-by-step transfer matrix for which the beta values should be computed
            twiss: the initial Twiss values
            plane: an integer representing the block (1 or 2)

        Returns:
            a Pandas Series with the gamma values computed at all steps of the input step-by-step transfer matrix
        """
        r11, r12, r21, r22, alpha, beta, gamma = _get_matrix_elements_block(m, twiss, plane)
        return r21 ** 2 * beta - 2.0 * r21 * r22 * alpha + r22 ** 2 * gamma

    @staticmethod
    def compute_mu_from_matrix(m: _pd.DataFrame, twiss: _Dict, plane: int = 1) -> _pd.Series:
        """
        Computes the phase advance values at every step of the input step-by-step transfer matrix.

        Args:
            m: the step-by-step transfer matrix for which the beta values should be computed
            twiss: the initial Twiss values
            plane: an integer representing the block (1 or 2)

        Returns:
            a Pandas Series with the phase advance computed at all steps of the input step-by-step transfer matrix
        """
        r11, r12, r21, r22, alpha, beta, gamma = _get_matrix_elements_block(m, twiss, plane)
        return _np.arctan2(r12.astype(_np.float64), (r11 * beta - r12 * alpha).astype(_np.float64))

    @staticmethod
    def compute_jacobian_from_matrix(m: _pd.DataFrame, twiss: _Dict, plane: int = 1) -> _pd.Series:
        """
        Computes the jacobian of the 2x2 transfer matrix (useful to verify the simplecticity).

        Args:
            m: the step-by-step transfer matrix for which the jacobians should be computed
            twiss: initial values for the Twiss parameters
            plane: an integer representing the block (1 or 2)

        Returns:
            a Pandas Series with the jacobian computed at all steps of the input step-by-step transfer matrix
        """
        r11, r12, r21, r22, alpha, beta, gamma = _get_matrix_elements_block(m, twiss, plane)
        return r11 * r22 - r12 * r21

    @staticmethod
    def compute_dispersion_from_matrix(m: _pd.DataFrame, twiss: _Dict, plane: int = 1) -> _pd.Series:
        """
        Computes the dispersion function at every step of the input step-by-step transfer matrix.

        Args:
            m: the step-by-step transfer matrix for which the dispersion function should be computed
            twiss: initial values for the Twiss parameters
            plane: an integer representing the block (1 or 2)

        Returns:
            a Pandas Series with the dispersion function computed at all steps of the input step-by-step transfer matrix

        """
        p = 1 if plane == 1 else 3
        if p == 1:
            d0 = twiss['DISP1']
            dp0 = twiss['DISP2']
        else:
            d0 = twiss['DISP3']
            dp0 = twiss['DISP4']
        r11: _pd.Series = m[f"R{p}{p}"]
        r12: _pd.Series = m[f"R{p}{p + 1}"]
        r15: _pd.Series = m[f"R{p}5"]
        return d0 * r11 + dp0 * r12 + r15

    @staticmethod
    def compute_dispersion_prime_from_matrix(m: _pd.DataFrame, twiss: _Dict, plane: int = 1) -> _pd.Series:
        """
        Computes the dispersion prime function at every step of the input step-by-step transfer matrix.

        Args:
            m: the step-by-step transfer matrix for which the dispersion prime function should be computed
            twiss: initial values for the Twiss parameters
            plane: an integer representing the block (1 or 2)

        Returns:
            a Pandas Series with the dispersion prime function computed at all steps of the input step-by-step transfer
            matrix

        Example:

        """
        p = 1 if plane == 1 else 3
        if p == 1:
            d0 = twiss['DISP1']
            dp0 = twiss['DISP2']
        else:
            d0 = twiss['DISP3']
            dp0 = twiss['DISP4']
        r21: _pd.Series = m[f"R{p + 1}{p}"]
        r22: _pd.Series = m[f"R{p + 1}{p + 1}"]
        r25: _pd.Series = m[f"R{p + 1}5"]
        return d0 * r21 + dp0 * r22 + r25

    @staticmethod
    def compute_periodic_twiss(matrix: _pd.DataFrame) -> _Dict:
        """
        Compute twiss parameters from a transfer matrix which is assumed to be a periodic transfer matrix.

        Args:
            matrix: the (periodic) transfer matrix

        Returns:
            a Series object with the values of the periodic Twiss parameters.
        """
        twiss = dict({
            'CMU1': (matrix['R11'] + matrix['R22']) / 2.0,
            'CMU2': (matrix['R33'] + matrix['R44']) / 2.0,
        })
        if twiss['CMU1'] < -1.0 or twiss['CMU1'] > 1.0:
            _warning(f"Horizontal motion is unstable; proceed with caution (cos(mu) = {twiss['CMU1']}).")
        with _warnings.catch_warnings():
            _warnings.simplefilter("ignore")
            twiss['MU1'] = _np.arccos(twiss['CMU1'])
        if twiss['CMU2'] < -1.0 or twiss['CMU2'] > 1.0:
            _warning(f"Vertical motion is unstable; proceed with caution (cos(mu) = {twiss['CMU2']}).")
        with _warnings.catch_warnings():
            _warnings.simplefilter("ignore")
            twiss['MU2'] = _np.arccos(twiss['CMU2'])
        twiss['BETA11'] = matrix['R12'] / _np.sin(twiss['MU1']) * _ureg.m
        if twiss['BETA11'] < 0.0:
            twiss['BETA11'] *= -1
            twiss['MU1'] *= -1
        twiss['BETA22'] = matrix['R34'] / _np.sin(twiss['MU2']) * _ureg.m
        if twiss['BETA22'] < 0.0:
            twiss['BETA22'] *= -1
            twiss['MU2'] *= -1
        twiss['ALPHA11'] = (matrix['R11'] - matrix['R22']) / (2.0 * _np.sin(twiss['MU1']))
        twiss['ALPHA22'] = (matrix['R33'] - matrix['R44']) / (2.0 * _np.sin(twiss['MU2']))
        twiss['GAMMA11'] = -matrix['R21'] / _np.sin(twiss['MU1']) * _ureg.m ** -1
        twiss['GAMMA22'] = -matrix['R43'] / _np.sin(twiss['MU2']) * _ureg.m ** -1
        m44 = matrix[['R11', 'R12', 'R13', 'R14',
                 'R21', 'R22', 'R23', 'R24',
                 'R31', 'R32', 'R33', 'R34',
                 'R41', 'R42', 'R43', 'R44']].apply(float).values.reshape(4, 4)
        r6 = matrix[['R15', 'R25', 'R35', 'R45']].apply(float).values.reshape(4, 1)
        disp = _np.dot(_np.linalg.inv(_np.identity(4) - m44), r6).reshape(4)
        twiss['DY'] = disp[0] * _ureg.m
        twiss['DYP'] = disp[1] * _ureg.radians
        twiss['DZ'] = disp[2] * _ureg.m
        twiss['DZP'] = disp[3] * _ureg.radians
        twiss['DISP1'] = twiss['DY']
        twiss['DISP2'] = twiss['DYP']
        twiss['DISP3'] = twiss['DZ']
        twiss['DISP4'] = twiss['DZP']

        return twiss
